#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Aug  7 09:10:20 2019

@author: davidblair
"""
import torch
import pyro
from pyro.infer import SVI
from pyro.infer import Trace_ELBO
from torch.utils import data
from pyro.optim import OneCycleLR
from torch.optim import AdamW
from collections import deque
import numpy as np
from scipy.stats import linregress
from pyro import poutine
from vlpi.data.ClinicalDataset import _TorchDatasetWrapper
from vlpi.utils.UtilityFunctions import rel_diff


class AnnealingScheduler:
    def _cosine_anneal(self,currentStepFrac):
        cos_vals = np.cos(np.pi * currentStepFrac) + 1.0
        return self.finalTemp + ((self.initTemp - self.finalTemp) / 2.0) * cos_vals

    def _linear_anneal(self,currentStepFrac):
        return (self.finalTemp - self.initTemp) * currentStepFrac + self.initTemp

    def __init__(self,initTemp,finalTemp,totalNumSteps,scheduleType):
        """
        Scheduler for KL-Anneling.

        Parameters
        ----------
        initTemp : float
            Initial temperature.
        finalTemp : float
            Final temperature.
        totalNumSteps : int
            Number of total steps for annealing
        scheduleType : str
            Type of scheduler. Must be 'linear' or 'cosine'.

        Returns
        -------
        None

        """

        assert initTemp>=0.0 and finalTemp >=0.0,"Annealing temperatures must be greater than 0.0"
        self.initTemp=initTemp
        self.finalTemp=finalTemp
        self.totalNumSteps=totalNumSteps
        if scheduleType=='linear':
            self._anneal_func=self._linear_anneal
        else:
            self._anneal_func=self._cosine_anneal

    def CurrentTemp(self,currentStep):
        frac = currentStep/self.totalNumSteps
        if frac>1.0:
            return self.finalTemp
        else:
            return self._anneal_func(frac)


class Optimizer:

    def _returnELBOSlope(self,elbo_deque):
        return abs(linregress(range(len(elbo_deque)),elbo_deque)[0]/np.mean(elbo_deque))

    def _localShuffle(self,dataSample):
        """
        shuffles dataSample locally on disk/gpu
        """

        if self.useCuda:
            new_index = torch.randperm(dataSample[0].shape[0],device = self.device)
            old_index = torch.arange(dataSample[0].shape[0],device = self.device)
        else:
            new_index = torch.randperm(dataSample[0].shape[0])
            old_index = torch.arange(dataSample[0].shape[0])

        dataSample[0][old_index] = dataSample[0][new_index]
        for i in range(len(dataSample[1])):
            dataSample[1][i][old_index]=dataSample[1][i][new_index]

        if dataSample[2] is not None:
            dataSample[2][old_index]=dataSample[2][new_index]
        if dataSample[3] is not None:
            dataSample[3][old_index]=dataSample[3][new_index]


    def _sendDataToGPU(self,returnData):
        """
        sends returnData to gpu
        """

        newIncData=returnData[0].to('cuda:{}'.format(self.device))
        newCovData=[]
        for i,arr in enumerate(returnData[1]):
           newCovData+=[arr.to('cuda:{}'.format(self.device))]

        if returnData[2] is not None:
            newTargetData=returnData[2].to('cuda:{}'.format(self.device))
        else:
            newTargetData=None
        if returnData[3] is not None:
            newScoreData=returnData[3].to('cuda:{}'.format(self.device))
        else:
            newScoreData=None

        return (newIncData,newCovData,newTargetData,newScoreData)




    def __init__(self,phenotypeModel,datasetSampler,optimizationParameters={'maxLearningRate': 0.05,'maxEpochs': 5000,'numParticles':10},computeConfiguration={'device':None,'numDataLoaders':0}, **kwargs):
        """

        This class implements two SGD optimizers, one of which uses the full training dataset (FullDatasetTrain) and the other uses mini-batches randomly sampled from the training dataset. To improve convergence rate and escape local minima, learning rate can be autommatically altered during inference. Note, given stochastic inference strategy coupled with possible re-starts, class tracks model perfomance and stores the best possible instance of the model obtained throughout the inference process. This also helps to avoid overfitting.

                phenotypeModel-->The phenotype model to optimize
                datasetSampler-->ClinicalDatasetSampler that holds clinical data and generates random samples

                maxLearningRate-->maximum learning rate used by SGD during inference.
                device-->compute device to use. By default, uses CPU unless GPU device specified, which can be string or integer

        Parameters
        ----------
        phenotypeModel : vlpi.model.VAE
            The phenotype model to optimize.
        datasetSampler : vlpi.data.ClinicalDatasetSampler
            ClinicalDatasetSampler that generates data
        optimizationParameters : dict
            Dictionary containing optimzation parameters. Default: {'maxLearningRate': 0.05,'maxEpochs': 5000,'numParticles':10}
        computeConfiguration : dict
            Dictionary containing the compute configuration (device and number of dataloaders. Defaults to using cpu: {'device':None,'numDataLoaders':0}
        **kwargs : See source code.
            These arguments adjust the parameters of the One-Cycle learning rate policy ('OneCycleParams'), Adam weight decay parameter ('AdamW_Weight_Decay'), and the KL-Annealing scheduler ('KLAnnealingParams'). See source code for details.

        Returns
        -------
        None

        """



        allKeywordArgs = list(kwargs.keys())

        self.phenotypeModel=phenotypeModel
        self.datasetSampler=datasetSampler

        #general optimization parameters
        self.maxEpochs=optimizationParameters['maxEpochs']
        self.maxLearningRate = optimizationParameters['maxLearningRate']
        self.numParticles = optimizationParameters['numParticles']

        #compute resources parameters
        self.device=computeConfiguration['device']
        if self.device is None:
            self.device='cpu'

        if self.device!='cpu':
            assert torch.cuda.is_available()==True, 'GPU is not available on your system.'
            self.useCuda=True
            if isinstance(self.device,str):
                self.device=int(self.device.strip('cuda:'))

        else:
            self.useCuda=False

        self.num_dataloaders=computeConfiguration['numDataLoaders']




        if 'OneCycleParams' not in allKeywordArgs:
            self.OneCycleParams={'pctCycleIncrease':1.0,'initLRDivisionFactor':1.0,'finalLRDivisionFactor':1.0}

        else:
            self.OneCycleParams=kwargs['OneCycleParams']

        if 'AdamW_Weight_Decay' in allKeywordArgs:
            self.AdamW_Weight_Decay=kwargs['AdamW_Weight_Decay']
        else:
            self.AdamW_Weight_Decay=1e-4

        if 'KLAnnealingParams' not in allKeywordArgs:
            self.KLAnnealingParams = {'initialTemp':1.0,'maxTemp':1.0,'fractionalDuration':1.0,'schedule': 'cosine'}
        else:
            self.KLAnnealingParams=kwargs['KLAnnealingParams']
            assert set(self.KLAnnealingParams.keys())==set(['initialTemp','maxTemp','fractionalDuration','schedule']),"KL Annealing Parameters must be dictionary with the following keys: 'initialTemp','maxTemp','fractionalDuration','schedule'"





    def _contstructSVI(self,optimizationStrategy,total_steps):
        """
        Constructs a list of SVI functions use to train model. Takes optimizationStrategy as input, which must be in ['Full','PosteriorOnly','GuideOnly']
        """

        #generate random sample just to build the param_store
        test_data = self.datasetSampler.GenerateRandomTrainingSample(2)
        if self.useCuda:
                test_data=self._sendDataToGPU(test_data)
        self.phenotypeModel.guide(*test_data)
        param_store=pyro.get_param_store()

        if optimizationStrategy=='Full':
            _model=self.phenotypeModel.model
            _guide=self.phenotypeModel.guide

        else:
            if optimizationStrategy=='DecoderOnly':
                _model=self.phenotypeModel.model
                _guide=poutine.block(self.phenotypeModel.guide,hide=[x for x in param_store.get_all_param_names() if ('encoder' in x)])

            else:
                _model=poutine.block(self.phenotypeModel.model,hide=[x for x in param_store.get_all_param_names() if ('decoder' in x)])
                _guide=self.phenotypeModel.guide

        #initialize the SVI class
        AdamWOptimArgs = {'weight_decay':self.AdamW_Weight_Decay}

        scheduler = OneCycleLR({'optimizer': AdamW, 'optim_args': AdamWOptimArgs,'max_lr':self.maxLearningRate,'total_steps':total_steps,'pct_start':self.OneCycleParams['pctCycleIncrease'],'div_factor':self.OneCycleParams['initLRDivisionFactor'],'final_div_factor':self.OneCycleParams['finalLRDivisionFactor']})
        return {'svi':SVI(_model,_guide,scheduler, loss=Trace_ELBO(num_particles=self.numParticles)),'scheduler':scheduler}


    def FullDatasetTrain(self, errorTol:float = 1e-3,verbose:bool=True,logFile=None,optimizationStrategy='Full',errorComputationWindow=None):
        """

        Trains the VAE model using the full dataset. Not recommended.

        Parameters
        ----------
        errorTol : float
            Error tolerance in validation data ELBO for convergence.
        verbose : bool
            Whether to print updates regarding fitting.
        logFile : str
            Path to logfile which can store progress
        optimizationStrategy : str
            Allows user to update only certain parts of model. Useful for debugging. Must be one of: 'Full','EncoderOnly','DecoderOnly'
        errorComputationWindow : float
            Sliding window for computing error tolerance. Default appears to work well.

        Returns
        -------
        tuple
            (bestModelScore, vector of ELBO evaluations on training data,vector of ELBO evaluations on testing data, vector of all errors computed during fitting)

        """

        if errorComputationWindow is None:
            errorComputationWindow = max(int(0.1*self.maxEpochs),2)
        else:
            assert errorComputationWindow>0, "Expects errorComputationWindow to be integer > 0"

        error_window = deque(maxlen=errorComputationWindow)
        elbo_window = deque(maxlen=errorComputationWindow)
        errorVec = []
        trainLoss = []
        testLoss = []



        if self.useCuda:
            self.phenotypeModel.SwitchDevice(self.device)

        self.phenotypeModel.train()

        #training and test data are fixed for full dataset training; therefore, just need
        #load into memory once.
        if self.useCuda:
            trainData=self._sendDataToGPU(self.datasetSampler.ReturnFullTrainingDataset())
            testData=self._sendDataToGPU(self.datasetSampler.ReturnFullTestingDataset())
        else:
            trainData=self.datasetSampler.ReturnFullTrainingDataset()
            testData=self.datasetSampler.ReturnFullTestingDataset()



        assert optimizationStrategy in ['Full','EncoderOnly','DecoderOnly'],"Available Optimization strategies: ['Full','EncoderOnly','DecoderOnly']"
        sviFunction = self._contstructSVI(optimizationStrategy,self.maxEpochs)
        annealer =AnnealingScheduler(self.KLAnnealingParams['initialTemp'],self.KLAnnealingParams['maxTemp'],int(self.maxEpochs*self.KLAnnealingParams['fractionalDuration']),self.KLAnnealingParams['schedule'])
        paramUpdateNum = 0


        # initialize model and parameter storage vectors. Note, parameter storage vectors are stored in a list with each element corresponding to one SVI function
        bestModelState= self.phenotypeModel.PackageCurrentState()
        bestModelScore = sviFunction['svi'].evaluate_loss(*testData,annealing_factor=annealer.CurrentTemp(paramUpdateNum))
        prev_train_loss = sviFunction['svi'].evaluate_loss(*trainData,annealing_factor=self.KLAnnealingParams['maxTemp'])
        elbo_window.append(prev_train_loss)

        for epoch in range(self.maxEpochs):
            self._localShuffle(trainData)

            # take step on training data

            paramUpdateNum+=1
            currrent_train_loss=sviFunction['svi'].step(*trainData,annealing_factor=annealer.CurrentTemp(paramUpdateNum))

            #adjust learning rate per scheduler
            sviFunction['scheduler'].step()


            #no point in shuffling testing data, all observations are independent
            #also, all svi functions should return identical loss, so we arbitrarily choose the latter for loss
            current_test_loss=sviFunction['svi'].evaluate_loss(*testData,annealing_factor=self.KLAnnealingParams['maxTemp'])

            if current_test_loss<bestModelScore:
                bestModelState = self.phenotypeModel.PackageCurrentState()
                bestModelScore = current_test_loss

            #track overall convergence
            trainLoss+=[currrent_train_loss]
            testLoss+=[current_test_loss]

            error_window.append(rel_diff(currrent_train_loss,prev_train_loss))
            elbo_window.append(currrent_train_loss)

            avg_error = sum(error_window)/len(error_window)
            med_error = sorted(error_window)[int(0.5*len(error_window))]
            slope_error = self._returnELBOSlope(elbo_window)
            prev_train_loss=currrent_train_loss

            errorVec+=[min([avg_error,med_error,slope_error])]

            if (avg_error < errorTol) or (med_error < errorTol) or (slope_error<errorTol):


                if verbose:
                    print("Optimization converged in %03d epochs; Current Loss (Train, Test): %.4f, %.4f; Error: %.4e"%(epoch+1,trainLoss[-1],testLoss[-1],min([avg_error,med_error,slope_error])))
                if logFile!=None:
                    with open(logFile, "a") as f:
                        f.write("Optimization converged in %03d epochs; Current Loss (Train, Test): %.4f, %.4f; Error: %.4e\n"%(epoch+1,trainLoss[-1],testLoss[-1],min([avg_error,med_error,slope_error])))
                break

            else:
                if verbose:
                    print("Completed %03d epochs; Current Loss (Train, Test): %.4f, %.4f; Error: %.4e"%(epoch+1,trainLoss[-1],testLoss[-1],min([avg_error,med_error,slope_error])))
                if logFile!=None:
                    with open(logFile, "a") as f:
                        f.write("Completed %03d epochs; Current Loss (Train, Test): %.4f, %.4f; Error: %.4e\n"%(epoch+1,trainLoss[-1],testLoss[-1],errorVec[-1][0],min([avg_error,med_error,slope_error])))


        if epoch == (self.maxEpochs-1):
            print('Warning: Optimization did not converge in allotted epochs')

        self.phenotypeModel.LoadPriorState(bestModelState)
        #set eval mode
        if self.useCuda:
            self.phenotypeModel.SwitchDevice('cpu')
        self.phenotypeModel.eval()
        return bestModelScore,trainLoss,testLoss,errorVec

    def BatchTrain(self,batch_size:int,errorTol:float = 1e-3,verbose=True,logFile=None,optimizationStrategy='Full',errorComputationWindow=None):
        """

        Trains the VAE model using the mini-batches. This is the recommended method.

        Parameters
        ----------
        batch_size : int
            Number of samples in each mini-batch
        errorTol : float
            Error tolerance in validation data ELBO for convergence.
        verbose : bool
            Whether to print updates regarding fitting.
        logFile : str
            Path to logfile which can store progress
        optimizationStrategy : str
            Allows user to update only certain parts of model. Useful for debugging. Must be one of: 'Full','EncoderOnly','DecoderOnly'
        errorComputationWindow : float
            Sliding window for computing error tolerance. Default appears to work well.

        Returns
        -------
        tuple
            (bestModelScore, vector of ELBO evaluations on training data,vector of ELBO evaluations on testing data, vector of all errors computed during fitting)

        """
        if errorComputationWindow is None:
            errorComputationWindow = max(int(0.05*self.maxEpochs),2)
        else:
            assert errorComputationWindow>0, "Expects errorComputationWindow to be integer > 0"

        error_window = deque(maxlen=errorComputationWindow)
        elbo_window = deque(maxlen=errorComputationWindow)
        errorVec = []
        trainLoss = []
        testLoss = []

        if isinstance(self.datasetSampler.trainingDataIndex,list):
            numTotalTrainingSamples=sum([len(x) for x in self.datasetSampler.trainingDataIndex])
            numTotalTestingSamples=sum([len(x) for x in self.datasetSampler.testDataIndex])
        else:
            numTotalTrainingSamples=len(self.datasetSampler.trainingDataIndex)
            numTotalTestingSamples=len(self.datasetSampler.testDataIndex)

        assert batch_size < numTotalTrainingSamples, "Batch size is greater than or equal to training data size. Use FullDatasetTrain"



        if self.useCuda:
            self.phenotypeModel.SwitchDevice(self.device)

        self.phenotypeModel.train()


        # for batch training, use PyTorch dataset loaders, as they allow threaded queuing of samples, which can greatly improve performance.


        torchTrainingData=_TorchDatasetWrapper(self.datasetSampler,self.datasetSampler.trainingDataIndex,batch_size)
        torchTestingData=_TorchDatasetWrapper(self.datasetSampler,self.datasetSampler.testDataIndex,batch_size)



        assert optimizationStrategy in ['Full','EncoderOnly','DecoderOnly'],"Available Optimization strategies: ['Full','EncoderOnly','DecoderOnly']"
        sviFunction = self._contstructSVI(optimizationStrategy,len(torchTrainingData)*self.maxEpochs)
        annealer=AnnealingScheduler(self.KLAnnealingParams['initialTemp'],self.KLAnnealingParams['maxTemp'],int(len(torchTrainingData)*(self.maxEpochs*self.KLAnnealingParams['fractionalDuration'])),self.KLAnnealingParams['schedule'])
        paramUpdateNum=0


        #initialize model states and scores
        bestModelState= self.phenotypeModel.PackageCurrentState()

        #note, batch_size handled at the level of _TorchDatasetWrapper, not loader itself.
        trainingDataLoader = data.DataLoader(torchTrainingData,batch_size=1,shuffle=False,sampler=None,batch_sampler=None,num_workers=self.num_dataloaders,collate_fn=lambda x:x[0],pin_memory=self.useCuda)
        testingDataLoader = data.DataLoader(torchTestingData,batch_size=1,shuffle=False,sampler=None,batch_sampler=None,num_workers=self.num_dataloaders,collate_fn=lambda x:x[0],pin_memory=self.useCuda)

        avg_epoch_train_loss = 0.0
        for i,data_batch in enumerate(trainingDataLoader):
            if self.useCuda:
                data_batch=self._sendDataToGPU(data_batch)

            avg_epoch_train_loss+=sviFunction['svi'].evaluate_loss(*data_batch,minibatch_scale = (numTotalTrainingSamples/data_batch[0].shape[0]),annealing_factor=annealer.CurrentTemp(paramUpdateNum+1))
        prev_train_loss=avg_epoch_train_loss/(i+1)

        elbo_window.append(prev_train_loss)

        avg_epoch_test_loss = 0.0

        for i,data_batch in enumerate(testingDataLoader):
            if self.useCuda:
                data_batch=self._sendDataToGPU(data_batch)
            avg_epoch_test_loss+=sviFunction['svi'].evaluate_loss(*data_batch,minibatch_scale = (numTotalTestingSamples/data_batch[0].shape[0]),annealing_factor=self.KLAnnealingParams['maxTemp'])
        bestModelScore = avg_epoch_test_loss/(i+1)


        for epoch in range(self.maxEpochs):
            torchTrainingData.shuffle_index()
            avg_epoch_train_loss = 0.0
            for i,data_batch in enumerate(trainingDataLoader):
                paramUpdateNum+=1
                if self.useCuda:
                    data_batch=self._sendDataToGPU(data_batch)
                avg_epoch_train_loss+=sviFunction['svi'].step(*data_batch,minibatch_scale = (numTotalTrainingSamples/data_batch[0].shape[0]),annealing_factor=annealer.CurrentTemp(paramUpdateNum))
                sviFunction['scheduler'].step()


            avg_epoch_train_loss=avg_epoch_train_loss/(i+1.0)

            avg_epoch_test_loss = 0.0
            for i,data_batch in enumerate(testingDataLoader):
                if self.useCuda:
                    data_batch=self._sendDataToGPU(data_batch)
                avg_epoch_test_loss+=sviFunction['svi'].evaluate_loss(*data_batch,minibatch_scale = (numTotalTestingSamples/data_batch[0].shape[0]),annealing_factor=self.KLAnnealingParams['maxTemp'])
            avg_epoch_test_loss=avg_epoch_test_loss/(i+1)

            if np.isnan(avg_epoch_test_loss) or np.isnan(avg_epoch_train_loss):
                print("Warning: NaN detected during inference. Model unlikely to be fully optimized!")
                break

            elif avg_epoch_test_loss<bestModelScore:
                bestModelState = self.phenotypeModel.PackageCurrentState()
                bestModelScore = avg_epoch_test_loss


            #track overall convergence
            trainLoss+=[avg_epoch_train_loss]
            testLoss+=[avg_epoch_test_loss]



            error_window.append(rel_diff(avg_epoch_train_loss,prev_train_loss))
            elbo_window.append(avg_epoch_train_loss)

            avg_error = sum(error_window)/len(error_window)
            med_error = sorted(error_window)[int(0.5*len(error_window))]
            slope_error = self._returnELBOSlope(elbo_window)

            prev_train_loss=avg_epoch_train_loss
            errorVec+=[min([avg_error,med_error,slope_error])]

            if (avg_error < errorTol) or (med_error < errorTol) or (slope_error<errorTol):


                if verbose:
                    print("Optimization converged in %03d epochs; Current Loss (Train, Test): %.4f, %.4f; Error: %.4e"%(epoch+1,trainLoss[-1],testLoss[-1],min([avg_error,med_error,slope_error])))
                if logFile!=None:
                    with open(logFile, "a") as f:
                        f.write("Optimization converged in %03d epochs; Current Loss (Train, Test): %.4f, %.4f; Error: %.4e\n"%(epoch+1,trainLoss[-1],testLoss[-1],min([avg_error,med_error,slope_error])))
                break

            else:
                if verbose:
                    print("Completed %03d epochs; Current Loss (Train, Test): %.4f, %.4f; Error: %.4e"%(epoch+1,trainLoss[-1],testLoss[-1],min([avg_error,med_error,slope_error])))
                if logFile!=None:
                    with open(logFile, "a") as f:
                        f.write("Completed %03d epochs; Current Loss (Train, Test): %.4f, %.4f; Error: %.4e\n"%(epoch+1,trainLoss[-1],testLoss[-1],errorVec[-1][0],min([avg_error,med_error,slope_error])))


        if epoch == (self.maxEpochs-1):
            print('Warning: Optimization did not converge in allotted epochs')

        self.phenotypeModel.LoadPriorState(bestModelState)
        #set eval mode
        if self.useCuda:
            self.phenotypeModel.SwitchDevice('cpu')
        self.phenotypeModel.eval()
        return bestModelScore,trainLoss,testLoss,errorVec



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Jul 16 15:37:15 2019

@author: davidblair
"""

import numpy as np
import pandas as pd
import torch
from torch.utils import data
from scipy import sparse
from typing import Iterable
from collections import OrderedDict
from sklearn.utils import shuffle
import copy
import pickle
import itertools
from vlpi.utils.UtilityFunctions import one_hot_scipy,one_hot
from vlpi.data.ICDUtilities import ICDUtilities

class ClinicalDataset:

    def none_to_int(self,val):
        if val!=None:
            return 1
        else:
            return 0

    def initialize_empty_df(self,columns, dtypes, index=None):
        assert len(columns)==len(dtypes)
        df = pd.DataFrame(index=index)
        for c,d in zip(columns, dtypes):
            df[c] = pd.Series(dtype=d)
        return df

    def _parseICDCodeList(self,codeList):
        codeList = list(set(codeList.strip().split(',')))
        codeList=[x.strip() for x in codeList if x.strip()!='']

        newCodeList=[]
        for code in codeList:
            try:
                newCodeList+=[self.dxCodeToDataIndexMap[code]]
            except KeyError:
                pass

        return newCodeList

    def _parseCatCov(self,covList,catCovNames):
        covVals = []
        for i,covName in enumerate(catCovNames):
            try:
                covVals+=[(covName,self.catCovConversionDicts[covName][covList[i]])]
            except KeyError:
                newIndex = len(self.catCovConversionDicts[covName])
                self.catCovConversionDicts[covName][covList[i]]=newIndex
                covVals+=[(covName,newIndex)]
        return covVals


    def __init__(self,ICDFilePaths:Iterable[str]=[]):
        """


        Parameters
        ----------
        ICDFilePaths : Iterable[str], optional
            This passes a list of strings ([ICD_hierarchy, ICD_chapters], see ICDUtilities) in order to initialize Dx Code data structure and mappings. This is only relevant when constructing new datasets from a flat text file. Otherwise, the Dx Code information is read from the stored ClinicalDataset Object, so the file paths are irrelevant. By default, the class instantiates the 2018 ICD10-CM coding structure, which is included with the software (as is the UKBB ICD10 encoding structure, downloaded in Jan 2020).

            The default is value [], which defaults to ICD10-CM 2018.

        Returns
        -------
        None.

        """
        if len(ICDFilePaths)==0:
            self.ICDInfo=ICDUtilities()
        else:
            assert len(ICDFilePaths)==2, "Expects a list containing 2 elecments: file paths for ICD10 hierarchy and chapters"
            self.ICDInfo=ICDUtilities(hierarchyFile=ICDFilePaths[0],chapterFile=ICDFilePaths[1])

        #initialize the clinical data structure to line up with the ICD codebook,
        #although the clinical dataset need not correspond strictly to ICD codes (and this is frequently true)
        self.dxCodeToDataIndexMap = copy.deepcopy(self.ICDInfo.usableCodeToIndexMap)
        self.dataIndexToDxCodeMap = dict(zip(self.dxCodeToDataIndexMap.values(),self.dxCodeToDataIndexMap.keys()))

        self.numDxCodes = len(self.dxCodeToDataIndexMap)
        self.data=None
        self.numPatients = None
        self.catCovConversionDicts={}

    def ReadDatasetFromFile(self,clinicalDataset,dxCodeColumn,indexColumn = None, skipColumns=[], hasHeader=True,chunkSize = 500):
        """

        Initializes the Pandas clinical dataset by reading it from a text file.

        Expects that clinical dataset is in ICD format. Can transition to other formats (HPO)
        by using using ConvertCodes function.

        Parameters
        ----------
        clinicalDataset : str
            File Name for clinical dataset.
        dxCodeColumn : int
            Column that contains a comma-separated list of associated ICD codes, first column denoted by 0
        indexColumn : int
            Column to use as index for the dataset
        skipColumns : list of ints
            List that indicates which columns should be skipped [uses 0-based indexing]
        hasHeader : type
            Indicates whether file has header, which is used to generate column names
        chunkSize : type
            Indicates how often database should be written into. Defaults to every 500 lines.

        Returns
        -------
        None

        """

        assert chunkSize >1, "chunkSize must be > 1"
        clinicalFile = open(clinicalDataset)
        if hasHeader:
            headLine = clinicalFile.readline().strip('\n').split('\t')
            catCovNames = [h for h in headLine if headLine.index(h) not in [dxCodeColumn,indexColumn]+skipColumns]
        else:
            currentLine = clinicalFile.readline().strip('\n').split('\t')
            catCovNames=['Covariate_'+str(i+1) for i in range(len(currentLine)-(1+self.none_to_int(indexColumn)+len(skipColumns)))]

        colNames = ['patient_id','dx_codes']+catCovNames

        self.catCovConversionDicts = {covName:{} for covName in catCovNames}
        self.data = self.initialize_empty_df(colNames,[np.int64,object]+[np.int32 for i in range(len(catCovNames))])
        patientCounter = int(0)
        currentDataList ={colName:[] for colName in self.data.columns}
        for line in clinicalFile:
            line = line.strip('\n').split('\t')
            currentDataList['dx_codes']+=[self._parseICDCodeList(line[dxCodeColumn])]
            for nm, val in self._parseCatCov([line[i] for i in range(len(line)) if i not in [dxCodeColumn,indexColumn]+skipColumns],catCovNames):
                currentDataList[nm]+=[val]

            if indexColumn!=None:
                currentDataList['patient_id']+=[int(line[indexColumn])]
            else:
                currentDataList['patient_id']+=[patientCounter]
            patientCounter+=1
            if patientCounter % chunkSize==0:
                self.data=self.data.append(pd.DataFrame(currentDataList),ignore_index=True)
                currentDataList ={colName:[] for colName in self.data.columns}


        if len(currentDataList['patient_id'])>0:
            self.data=self.data.append(pd.DataFrame(currentDataList),ignore_index=True)

        #shuffle data and create new index
        self.data = self.data.sample(frac=1).reset_index(drop=True)
        self.data.set_index('patient_id',drop=False, inplace=True)
        self.numPatients = len(self.data)


    def FindAllPatients_wDx(self,dx_code):
        """

        Finds all patients with a particular dx code, returns their index vals.

        Parameters
        ----------
        dx_code : str
            ICD code string.

        Returns
        -------
        pd.Series
            Series containing index of patients with particular diagnosis.

        """
        if '.' in dx_code:
            dx_code.replace('.','')
        intVal = self.dxCodeToDataIndexMap[dx_code]
        return self.data['patient_id'][self.data['dx_codes'].apply(lambda x: intVal in x)]


    def IncludeOnly(self,dx_code_list):
        """
        Removes all dx_codes from the dataset except those from the dx_code_list.

        Parameters
        ----------
        dx_code_list : list of str
            List of ICD10 strings to include.

        Returns
        -------
        None

        """


        dx_code_list=[x.replace('.','') for x in dx_code_list]
        allKept = set([self.dxCodeToDataIndexMap[x] for x in dx_code_list])

        #now we need to remove all non-kept codes from the ICD conversion dictionaries
        removedCodes = set(self.dxCodeToDataIndexMap.keys()).difference(dx_code_list)

        for old_code in removedCodes:
            del self.dxCodeToDataIndexMap[old_code]

        self.dataIndexToDxCodeMap = {}
        newCodeToIntMap = {}
        oldToNewIntMap={}
        for i,key in enumerate(self.dxCodeToDataIndexMap):
            oldToNewIntMap[self.dxCodeToDataIndexMap[key]]=i
            self.dataIndexToDxCodeMap[i]=key
            newCodeToIntMap[key] = i

        self.dxCodeToDataIndexMap=newCodeToIntMap
        if isinstance(self.data,pd.DataFrame):
            self.data['dx_codes']=self.data['dx_codes'].apply(lambda x: [oldToNewIntMap[y] for y in x if y in allKept])

        self.numDxCodes=len(self.dxCodeToDataIndexMap)

    def ConvertCodes(self,dx_code_list:Iterable[str],new_code:str):
        """

        Converts set of ICD codes into a single dx code through logical-OR function. If given a single code, simply renames code as new_code.

        Parameters
        ----------
        dx_code_list : Iterable[str]
            List of codes to convert into a single, new code.
        new_code : str
            Name of new code

        Returns
        -------
        None
        """

        assert len(dx_code_list)>0, "dx_code_list must have one elemenent to collapse."
        dx_code_list=[x.replace('.','') for x in dx_code_list]

        #set the all codes in the list to the integer value of the first code in the list
        removedCodes = set(dx_code_list[1:])
        oldInts =[]
        for old_code in removedCodes:
            oldInts+=[self.dxCodeToDataIndexMap[old_code]]
            del self.dxCodeToDataIndexMap[old_code]


        if len(removedCodes)>0:
            self.dataIndexToDxCodeMap = {}
            newCodeToIntMap = {}
            oldToNewIntMap={}
            for i,key in enumerate(self.dxCodeToDataIndexMap):
                oldToNewIntMap[self.dxCodeToDataIndexMap[key]]=i
                self.dataIndexToDxCodeMap[i]=key
                newCodeToIntMap[key] = i

            self.dxCodeToDataIndexMap=newCodeToIntMap
            newInt = self.dxCodeToDataIndexMap[dx_code_list[0]]
            collapsedInt_to_Int = dict(zip(oldInts,[newInt for i in range(len(oldInts))]))
            oldToNewIntMap.update(collapsedInt_to_Int)
            if isinstance(self.data,pd.DataFrame):
                self.data['dx_codes']=self.data['dx_codes'].apply(lambda x: list(set([oldToNewIntMap[y] for y in x])))
        else:
            newInt = self.dxCodeToDataIndexMap[dx_code_list[0]]


        #update the code information

        self.dataIndexToDxCodeMap[newInt] = new_code
        self.dxCodeToDataIndexMap[new_code] = newInt
        del self.dxCodeToDataIndexMap[dx_code_list[0]]
        self.numDxCodes=len(self.dxCodeToDataIndexMap)


    def ConstructNewDataArray(self,oldCodeToNewMap):
        """This function translates diagnostic codes and symptom data to a new encoding using the dictionary oldCodeToNewMap. The previous codes are provided as keys (strings), and the new codes as values (also strings). The names for the new codes will be changed automatically. Values can also be iterables such that the old code maps to multiple new ones. Any code not provided as a key in the input dictionary will be dropped from the dataset.

        Parameters
        ----------
        oldCodeToNewMap : dict
            Key,value pairs indicating translation of old codes to new.

        Returns
        -------
        None

        """
        allNewCodes = sorted(list(set().union(*oldCodeToNewMap.values())))
        newCodeToIntMap = dict(zip(allNewCodes,range(len(allNewCodes))))
        newIntToCodeMap = dict(zip(newCodeToIntMap.values(),newCodeToIntMap.keys()))
        if self.data is not None:
            def _convFunc(x):
                newDxSet=set([])
                for dx in x:
                    try:
                        newDxSet.update(oldCodeToNewMap[self.dataIndexToDxCodeMap[dx]])
                    except KeyError:
                        pass
                return list({newCodeToIntMap[x] for x in newDxSet})

            self.data['dx_codes'] = self.data['dx_codes'].apply(_convFunc)
        self.dataIndexToDxCodeMap = newIntToCodeMap
        self.dxCodeToDataIndexMap = newCodeToIntMap
        self.numDxCodes=len(self.dxCodeToDataIndexMap)

    def ExcludeAll(self,dx_code_list):
        """
        Removes all codes in dx_code_list from the dataset

        Parameters
        ----------
        dx_code_list : list
            List of diagnostic codes to drop from the dataset.

        Returns
        -------
        None

        """

        keptCodes = set(self.dxCodeToDataIndexMap.keys()).difference(dx_code_list)
        self.IncludeOnly(list(keptCodes))

    def ConditionOnDx(self,dx_code_list):
        """

        This function conditions the data table on whether a patient has each diagnosis in the list 'dx_code_list'. This is accomplished by finding all patients with each diagnosis in 'dx_code_list', then adding a column (boolean) to the data table indicating diagnostic status. Each column will be named 'has_DX_CODE', where DX_CODE is the diagnostic code being conditioned on. These codes are then removed from the symptom data table. This function is expecially useful for supervised learning, as it creates labels from diagnostic codes.

        Parameters
        ----------
        dx_code_list : list of st
            List of codes to on which to condition the dataset

        Returns
        -------
        None.

        """

        for dx_code in dx_code_list:
            dx_code.replace('.','')
            allPatients_wDx=self.FindAllPatients_wDx(dx_code)
            hasDx=np.zeros(self.numPatients,dtype=np.bool)
            self.data.insert(len(self.data.columns),'has_'+dx_code,hasDx)
            self.data.loc[allPatients_wDx,'has_'+dx_code]=True
        self.ExcludeAll(dx_code_list)



    def WriteToDisk(self,fileName):
        """

        Writes ClinicalDataset to disk. Recommended way to store data after parsing text file.

        Parameters
        ----------
        fileName : str
            Path to storage file.

        Returns
        -------
        None

        """

        if fileName[-4:]!='.pth':
            fileName+='.pth'
        with open(fileName,'wb') as f:
            pickle.dump(self.data,f,protocol=pickle.HIGHEST_PROTOCOL)
            pickle.dump(self.catCovConversionDicts,f,protocol=pickle.HIGHEST_PROTOCOL)
            pickle.dump(self.dxCodeToDataIndexMap,f,protocol=pickle.HIGHEST_PROTOCOL)
            pickle.dump(self.dataIndexToDxCodeMap,f,protocol=pickle.HIGHEST_PROTOCOL)
            pickle.dump(self.numDxCodes,f,protocol=pickle.HIGHEST_PROTOCOL)

    def ReadFromDisk(self,fileName):
        """

        Reads ClinicalDataset written with WriteToDisk. To load a previously processed dataset, you must instantiate a ClinicalDataset class, which can then be used to read the file.

        Parameters
        ----------
        fileName : str
            Path to storage file.

        Returns
        -------
        None

        """

        if fileName[-4:]!='.pth':
            fileName+='.pth'
        with open(fileName,'rb') as f:
            self.data = pickle.load(f)
            self.catCovConversionDicts = pickle.load(f)
            self.dxCodeToDataIndexMap = pickle.load(f)
            self.dataIndexToDxCodeMap = pickle.load(f)
            self.numDxCodes = pickle.load(f)

        self.numPatients = len(self.data)



    def LoadFromArrays(self,incidenceArray,covariateArrays,covariateNames,catCovDicts=None, arrayType = 'Numpy'):
        """

        Loads clinical dataset from array data, generally used for simulation purposes. However, could also be used to bypass the ICD10 structure and load custom binary datasets. Dataset would need to be manipulated ahead of time using ConvertCodes and IncludeOnly to obtain a dataset with the dimensions/labels. Input arrays must be Numpy or PyTorch tensors.

        Parameters
        ----------
        incidenceArray : np.array or torch.tensor
            Binary symptom array
        covariateArrays : list of numpy.array or torch.tensor
            List of categorical covariates, which contains one numpy.array/torch.tensor per covariate
        covariateNames : List of str
            List of names for covariates
        catCovDicts : list of dicts
            List of dictionaries (one for each covariate) that maps covariates to integer values. If not provided, this is done automatically
        arrayType : str
            Indictes the array type. Numpy arrays ['Numpy'] or pyTorch tensors ['Torch']. Default is Numpy.

        Returns
        -------
        None

        """

        assert arrayType in ['Numpy','Torch'], "Only Numpy arrarys or Torch tensors supported"
        if covariateArrays==None:
            covariateArrays=[]
        if covariateNames==None:
            covariateNames=[]
        assert len(covariateArrays)==len(covariateNames), "Number of covariate names does not match number of covariate arrays."
        assert incidenceArray.shape[1]==self.numDxCodes, "Dimension of incidence data does not match number of codes."

        if arrayType=='Torch':
            incidenceArray=incidenceArray.to('cpu').detach().numpy()
            covariateArrays=[x.to('cpu').detach().numpy().ravel() for x in covariateArrays]
        else:
            covariateArrays=[x.ravel() for x in covariateArrays]

        dataDict={}
        for i,name in enumerate(covariateNames):
            if catCovDicts == None:
                uniqueCats = list(set(covariateArrays[i]))
                self.catCovConversionDicts[name] = dict(zip(uniqueCats,list(range(len(uniqueCats)))))
                covariateArrays[i] = np.array([self.catCovConversionDicts[name][x] for x in covariateArrays[i]],dtype=np.int64)
            else:
                self.catCovConversionDicts[name]=catCovDicts[i]
                covariateArrays[i] = np.array([self.catCovConversionDicts[name][x] for x in covariateArrays[i]],dtype=np.int64)
            dataDict[name] = covariateArrays[i]

        dataDict['patient_id']=np.arange(incidenceArray.shape[0],dtype=np.int64)
        dataDict['dx_codes'] = [np.where(x==1)[0].tolist() for x in incidenceArray]

        self.data = pd.DataFrame(dataDict)
        self.data = self.data.sample(frac=1).reset_index(drop=True)
        self.data.set_index('patient_id',drop=False, inplace=True)
        self.numPatients = len(self.data)

    def ReturnSparseDataMatrix(self,index:Iterable[int]=[]):
        """

        Returns disease incidence array as sparse coo matrix. Takes optional index, which returns only data points contained within the index.

        Parameters
        ----------
        index : Iterable[int]
            Index of patients to include.

        Returns
        -------
        sparse.coo_matrix
            Sparse, binary array of diagnoses.

        """

        if len(index)==0:
            index = self.data.index
        y_inds = list(itertools.chain.from_iterable(self.data.loc[index]['dx_codes']))
        x_inds = list(itertools.chain.from_iterable([[i]*len(x) for i,x in enumerate(self.data.loc[index]['dx_codes'])]))
        return sparse.coo_matrix((np.ones((len(x_inds))),(x_inds,y_inds)),shape=(len(index),self.numDxCodes),dtype=np.float32)




class ClinicalDatasetSampler():


    def _numpyWrapper(self,x):
        if x.dtype == np.float32:
            if sparse.issparse(x):
                return x.toarray()
            else:
                return np.array(x,dtype=np.float32)

        elif x.dtype == np.float64:
            if sparse.issparse(x):
                return x.toarray()
            else:
                return np.array(x,dtype=np.float64)
        else:
            if sparse.issparse(x):
                return x.toarray()
            else:
                return np.array(x,dtype=np.int64)

    def _scipySparseWrapper(self,x):
        if x.dtype == np.float32:
            if sparse.issparse(x):
                return x.tocsr()
            else:
                return sparse.csr_matrix(x,dtype=np.float32)

        elif x.dtype==np.float64:
            if sparse.issparse(x):
                return x.tocsr()
            else:
                return sparse.csr_matrix(x,dtype=np.float64)
        else:
            if sparse.issparse(x):
                return x.tocsr()
            else:
                return sparse.csr_matrix(x,dtype=np.int64)

    def _torchWrapper(self,x):

        """
        Note, all torch floating point tensors are converted to 32-bits to
        ensure GPU compatibility.
        """

        if x.dtype==np.float32:
            if sparse.issparse(x):
                return torch.tensor(x.toarray(),dtype = torch.float32)
            else:
                return torch.tensor(x,dtype = torch.float32)

        elif x.dtype==np.float64:
            if sparse.issparse(x):
                return torch.tensor(x.toarray(),dtype = torch.float32)
            else:
                return torch.tensor(x,dtype = torch.float32)
        else:
            if sparse.issparse(x):
                return torch.tensor(x.toarray(),dtype = torch.long)
            else:
                return torch.tensor(x,dtype = torch.long)


    def __init__(self, currentClinicalDataset,trainingFraction,conditionSamplingOnDx:Iterable[str]=[],returnArrays='Numpy',shuffle=True):
        """

        Generates random samples from a clinical dataset. Samples can be generated unconditionially, or conditional on a patient having a particular dx. Note, that in the latter case, the dx will be removed from the dataset and included as a separate column in the data if not already done.

        Parameters
        ----------
        currentClinicalDataset : ClinicalDataset
            Instance of the class ClinicalDataset
        trainingFraction : type
            Fraction of dataset used for training. Must be between 0.0 and 1.0.
        conditionSamplingOnDx : Iterable[str]
            Allows sampling to be conditioned on a set of diagnoses such that at least one patient in every sample had at least one of the diagnoses in the set. Note: original dataset is modified.
        returnArrays : str
            Array type returned by the sampling. Can be 'Numpy', 'Sparse' or 'Torch'. In the case of Sparse arrays, incidence arrays are returned as csr matrices, 1-d covariate vectors default to COO format.
        shuffle : bool
            Indcates whether to shuffle the data prior to splitting into training and test sets. Defaults to True, only make False for very large datasets that have already been shuffled.

        Returns
        -------
        None

        """
        self.conditionSamplingOnDx=conditionSamplingOnDx
        if len(conditionSamplingOnDx)>0:
            self.isConditioned = True
        else:
            self.isConditioned = False
        self.currentClinicalDataset=currentClinicalDataset
        self._returnAuxData=False
        self._auxDataset=None
        self.trainingFraction = trainingFraction
        assert self.trainingFraction >0.0 and self.trainingFraction<1.0, "Fraction of dataset used for training must be between 0.0 and 1.0."
        self.fracWDx=0.0
        self.numTotalSamples = len(self.currentClinicalDataset.data)
        self.includedCovariates = self.currentClinicalDataset.catCovConversionDicts.keys()


        assert returnArrays in ['Numpy','Torch','Sparse'], "Only Numpy arrarys, Torch tensors, or Scipy.Sparse (csr) supported"

        self.returnArrays=returnArrays
        if returnArrays =='Numpy':
            self.arrayFunc = self._numpyWrapper
        elif returnArrays =='Torch':
            self.arrayFunc=self._torchWrapper
        else:
            self.arrayFunc=self._scipySparseWrapper




        if shuffle==True:
            self.currentClinicalDataset.data=self.currentClinicalDataset.data.sample(frac=1)

        if len(self.conditionSamplingOnDx)==0:
            self.currentClinicalDataset = currentClinicalDataset
            cutOffVal = int(np.floor(len(currentClinicalDataset.data)*self.trainingFraction))
            self.trainingDataIndex = currentClinicalDataset.data.index[0:cutOffVal]
            self.testDataIndex = currentClinicalDataset.data.index[cutOffVal:]
        else:
            conditionedColumns = set(['has_'+dx_code for dx_code in self.conditionSamplingOnDx])
            missingColumns  = conditionedColumns.difference(self.currentClinicalDataset.data.columns)

            if len(missingColumns)>0:
                self.currentClinicalDataset.ConditionOnDx([x.replace('has_','') for x in missingColumns])

            has_at_least_one_dx = np.array(np.sum(np.vstack([self.currentClinicalDataset.data['has_'+dx] for dx in self.conditionSamplingOnDx]),axis=0),dtype=np.bool)
            dataWithDx = self.currentClinicalDataset.data.index[has_at_least_one_dx>0]
            dataWithoutDx = self.currentClinicalDataset.data.index[has_at_least_one_dx==0]
            self.fracWDx = len(dataWithDx)/len(self.currentClinicalDataset.data)
            cutOffValWDx = int(np.floor(len(dataWithDx)*self.trainingFraction))
            cutOffValWoDx = int(np.floor(len(dataWithoutDx)*self.trainingFraction))

            self.trainingDataIndex=[dataWithDx[0:cutOffValWDx],dataWithoutDx[0:cutOffValWoDx]]
            self.testDataIndex=[dataWithDx[cutOffValWDx:],dataWithoutDx[cutOffValWoDx:]]

    def DropSamples(self,index_vals,dropFromFullDataset=True):

        """

        Parameters
        ----------
        index_vals : array
            Index values to drop from the dataset.

        dropFromFullDataset : boolean; default True
            Indicates whether to drop the samples from the full dataset rather than only the sampler. By default, drops from the full dataset to avoid cases where samples are returned because data is accessed outside of sampler.

        Returns
        -------
        None
        """

        #first remove samples from indices
        if isinstance(self.trainingDataIndex,list)==False:
            self.trainingDataIndex=np.setdiff1d(self.trainingDataIndex,index_vals)
            self.testDataIndex=np.setdiff1d(self.trainingDataIndex,index_vals)
        else:
            self.trainingDataIndex=[np.setdiff1d(ind,index_vals) for ind in self.trainingDataIndex]
            self.testDataIndex=[np.setdiff1d(ind,index_vals) for ind in self.testDataIndex]
        if dropFromFullDataset==False:
            print("WARNING: Samples dropped from ClinicalDatasetSampler are still in the ClinicalDataset. Therefore, they can be returned by methods that bypass the Sampler!")

        else:
            index_vals=self.currentClinicalDataset.data.index.intersection(index_vals)
            self.currentClinicalDataset.data.drop(index=index_vals,inplace=True)




    def ChangeArrayType(self,newArrayType):
        """

        Changes the return array type.

        Parameters
        ----------
        newArrayType : str
            Must be one of 'Numpy','Torch','Sparse'

        Returns
        -------
        None

        """

        assert newArrayType in ['Numpy','Torch','Sparse'], "Only Numpy arrarys, Torch tensors, or Scipy.Sparse (csr) supported"
        self.returnArrays=newArrayType
        if newArrayType =='Numpy':
            self.arrayFunc = self._numpyWrapper
        elif newArrayType =='Torch':
            self.arrayFunc=self._torchWrapper
        else:
            self.arrayFunc=self._scipySparseWrapper



    def WriteToDisk(self,fName):
        """
        Writes sampler to disk so that it can be re-instantiated for further use. This is important for using the same test/training set across multiple models.

        Parameters
        ----------
        fName : str
            Path to storage file.

        Returns
        -------
        None

        """

        if fName[-4:]!='.pth':
            fName+='.pth'
        currentSampler = OrderedDict()
        currentSampler['conditionSamplingOnDx']=self.conditionSamplingOnDx
        currentSampler['numTotalSamples'] = self.numTotalSamples
        currentSampler['trainingDataIndex']=self.trainingDataIndex
        currentSampler['testDataIndex']=self.testDataIndex
        currentSampler['trainingFraction']=self.trainingFraction
        currentSampler['fracWDx']=self.fracWDx


        with open(fName, 'wb') as f:
            pickle.dump(currentSampler,f)



    def ReadFromDisk(self,fName):
        """
        Reads sampler from disk. This is important for using the same test/training set across multiple models.

        Parameters
        ----------
        fName : str
            Path to storage file.

        Returns
        -------
        None

        """
        if fName[-4:]!='.pth':
            fName+='.pth'

        with open(fName, 'rb') as f:
            currentSampler = pickle.load(f)
        #assertions to make sure that sampler parameters  match up

        assert currentSampler['numTotalSamples'] == self.numTotalSamples, "Datasets used by saved and current sampler are different lengths! Suspect they are not referring to same dataset"
        assert currentSampler['conditionSamplingOnDx']==self.conditionSamplingOnDx, "Saved and current sampler are not conditioned on same dx"
        assert currentSampler['trainingFraction']==self.trainingFraction,"Saved and current sampler do not have the same training fraction"
        self.testDataIndex=currentSampler['testDataIndex']
        self.trainingDataIndex=currentSampler['trainingDataIndex']
        self.fracWDx=currentSampler['fracWDx']

    def ConvertToUnconditional(self):
        """

        Converts a previously conditional sampler to unconditional while keeping the same testing and training sets. This way, the conditional diagnosis is NOT returned with the symptom/covariate data. Note, if unconditional, disease that is conditioned on won't be part of the symptom data array.

        Returns
        -------
        None

        """

        assert len(self.conditionSamplingOnDx)!=0, "Sampler is already uncoditional and was never conditional to start."
        assert isinstance(self.trainingDataIndex,list) is True, "Sampler has already been converted to unconditional."
        self.trainingDataIndex=np.concatenate(self.trainingDataIndex)
        self.testDataIndex=np.concatenate(self.testDataIndex)
        self.isConditioned=False

    def RevertToConditional(self):
        """

        Reverts a previously unconditional sampler to conditional while keeping the same testing and training sets. This way, the conditional diagnosis is returned with the symptom/covariate data.

        Returns
        -------
        None

        """
        assert len(self.conditionSamplingOnDx)!=0, "Sampler was not constructed as a conditional sampler. If you want a conditional sampler for this dataset, create a new ClincalDatasetSampler instance."
        assert isinstance(self.trainingDataIndex,list) is False, "Sampler is already conditional."
        has_at_least_one_dx_train= np.array(np.sum(np.vstack([self.currentClinicalDataset.data.loc[self.trainingDataIndex]['has_'+dx] for dx in self.conditionSamplingOnDx]),axis=0),dtype=np.bool)
        has_at_least_one_dx_test= np.array(np.sum(np.vstack([np.array(self.currentClinicalDataset.data.loc[self.testDataIndex]['has_'+dx],dtype=np.bool) for dx in self.conditionSamplingOnDx]),axis=0),dtype=np.bool)
        self.trainingDataIndex=[self.trainingDataIndex[has_at_least_one_dx_train],self.trainingDataIndex[np.invert(has_at_least_one_dx_train)]]
        self.testDataIndex=[self.testDataIndex[has_at_least_one_dx_test],self.testDataIndex[np.invert(has_at_least_one_dx_test)]]
        self.isConditioned=True

    def SubsetCovariates(self,newCovList):
        """
        Indicates the covariates contained within ClinicalDataset that should be returned by the sampler. Can be empty list, which indicates that no covariates should be returned.

        Parameters
        ----------
        newCovList : list
            List of covariate names

        Returns
        -------
        None.

        """
        assert set(newCovList).issubset(self.includedCovariates), "Subset of covariates provided is not subset  of current covariates."
        self.includedCovariates=newCovList


    def _returnData(self,newIndex,collapseAnchorDx=True):
        if isinstance(newIndex,Iterable)==False:
            newIndex=[newIndex]

        incidenceData = self.arrayFunc(self.currentClinicalDataset.ReturnSparseDataMatrix(newIndex))
        covData = [self.currentClinicalDataset.data.loc[newIndex][s].values for s in self.includedCovariates]
        covData = [self.arrayFunc(x.reshape(incidenceData.shape[0],1)) for x in covData]

        if not self.isConditioned:
            target_data = None
        else:
            target_data = np.array(pd.concat([self.currentClinicalDataset.data.loc[newIndex]['has_'+dx] for dx in self.conditionSamplingOnDx],axis=1),dtype=np.float32)
            target_data=self.arrayFunc(target_data.reshape(incidenceData.shape[0],len(self.conditionSamplingOnDx)))


        if not self._returnAuxData:
            encoded_data = None
        else:
            encoded_data = self.arrayFunc(self._auxDataset.ReturnSparseDataMatrix(newIndex))

        return incidenceData,covData,target_data,encoded_data

    def _generateRandomSample(self,numSamples,datasetIndex,fixedFracWDx):
        if fixedFracWDx!=None:
            assert isinstance(fixedFracWDx, float) and fixedFracWDx < 1.0 and fixedFracWDx > 0.0, "fixedFrac with dx must be float between 0.0 and 1.0"
            assert len(self.conditionSamplingOnDx)>0, "Cannot include fixed fraction of positive cases if conditionSamplingOnDx not enabled"
            if fixedFracWDx*numSamples >= len(datasetIndex[0]):
                numSamplesWDx = len(datasetIndex[0])
                print("Warning: fixedFracWDx exceeds or is equal to the total number of positive training samples.\nEvery positive case will be included in random sample")
            else:
                numSamplesWDx = int(np.ceil(numSamples*fixedFracWDx))
        elif self.isConditioned:
            numSamplesWDx=int(np.ceil(numSamples*self.fracWDx))


        if not self.isConditioned:
            newIndex = np.random.choice(datasetIndex,size=numSamples,replace=False)
        else:
            newIndex = shuffle(np.concatenate((np.random.choice(datasetIndex[0],size=numSamplesWDx,replace=False),np.random.choice(datasetIndex[1],size=(numSamples-numSamplesWDx),replace=False))))
        return self._returnData(newIndex)


    def GenerateRandomTrainingSample(self,numSamples, fixedFracWDx=None):
        """

        Returns a random subset of numSamples from training dataset.

        Parameters
        ----------
        numSamples : int
            Number of samples to return
        fixedFracWDx : float in [0.0,1.0]
            If the sampler is conditioned, will return a sample with fixedFracWDx*100% of subjects having the conditioned dx.

        Returns
        -------
        Tuple of arrays: (symptom data,list of covariate data, conditioned disease value, encoded data)

        """

        return self._generateRandomSample(numSamples,self.trainingDataIndex,fixedFracWDx)


    def GenerateRandomTestSample(self,numSamples,fixedFracWDx=None):
        """

        Returns a random subset of numSamples from testing dataset.

        Parameters
        ----------
        numSamples : int
            Number of samples to return
        fixedFracWDx : float in [0.0,1.0]
            If the sampler is conditioned, will return a sample with fixedFracWDx*100% of subjects having the conditioned dx.

        Returns
        -------
        Tuple of arrays: (symptom data,list of covariate data, conditioned disease value [if indicated], auxillary data [if included])

        """

        return self._generateRandomSample(numSamples,self.testDataIndex,fixedFracWDx)

    def _returnFullDataset(self,datasetIndex,randomize):
        if self.isConditioned:
            datasetIndex = np.concatenate(datasetIndex,axis=0)
        if randomize==True:
            datasetIndex=shuffle(datasetIndex)

        return self._returnData(datasetIndex)


    def ReturnFullTrainingDataset(self,randomize=True):
        """

        Returns the full training dataset.

        Returns
        -------
        Tuple of arrays: (symptom data,list of covariate data, conditioned disease value [if indicated], auxillary data [if included])

        """
        return self._returnFullDataset(self.trainingDataIndex,randomize)

    def ReturnFullTestingDataset(self,randomize=True):
        """

        Returns the full testing dataset.

        Returns
        -------
        Tuple of arrays: (symptom data,list of covariate data, conditioned disease value [if indicated], auxillary data [if included])
        """
        return self._returnFullDataset(self.testDataIndex,randomize)


    def _indexSplit(self,dataSetSize,totalNumBatches):
        if totalNumBatches>0:
            nEachBatch, extras = divmod(dataSetSize, totalNumBatches)
            section_sizes = ([0] +extras * [nEachBatch+1] +(totalNumBatches-extras) * [nEachBatch])
            return np.array(section_sizes, dtype=np.int32).cumsum()
        else:
            return np.array([0]+[dataSetSize], dtype=np.int32)


    def _epoch(self,datasetIndex,batch_size):

        if self.isConditioned:
            #shuffle the order of the dataset at the start of every epoch
            datasetIndex[0]=np.random.permutation(datasetIndex[0])
            datasetIndex[1]=np.random.permutation(datasetIndex[1])
            totalNumBatches,leftover = divmod((len(datasetIndex[0])+len(datasetIndex[1])),batch_size)
            assert totalNumBatches <= len(datasetIndex[0]), "Batch size too small. Cannot ensure at least one positive example per batch."
            #first shuffle the data

            hasDxSplits=self._indexSplit(len(datasetIndex[0]),totalNumBatches)
            noDxSplits=self._indexSplit(len(datasetIndex[1]),totalNumBatches)
            if totalNumBatches == 0:
                totalNumBatches+=1

            for i in range(totalNumBatches):
                hasDxSubset = datasetIndex[0][hasDxSplits[i]:hasDxSplits[i+1]]
                noDxSubset = datasetIndex[1][noDxSplits[i]:noDxSplits[i+1]]

                batchIndex = np.concatenate((hasDxSubset,noDxSubset))
                #need to shuffle positive cases within all others
                yield shuffle(batchIndex)
        else:
            datasetIndex=shuffle(datasetIndex)
            totalNumBatches,leftover = divmod(len(datasetIndex),batch_size)
            splits = self._indexSplit(len(datasetIndex),totalNumBatches)
            if totalNumBatches == 0:
                totalNumBatches+=1
            for i in range(totalNumBatches):

                yield datasetIndex[splits[i]:splits[i+1]]


    def TrainingEpochGenerator(self, batch_size):
        """
        Provides an iterator over the training dataset. Equivalent to performing one pass (epoch) through the dataset.

        Parameters
        ----------
        batch_size : int
            Batch size for the samples in the epoch.

        Returns
        -------
        iterator
            Iterates through training data samples.
        """
        for batch in self._epoch(self.trainingDataIndex,batch_size):
            yield self._returnData(batch)


    def TestingEpochGenerator(self,batch_size):
        """
        Provides an iterator over the testing dataset. Equivalent to performing one pass (epoch) through the dataset.

        Parameters
        ----------
        batch_size : int
            Batch size for the samples in the epoch.

        Returns
        -------
        iterator
            Iterates through testing data samples.
        """
        for batch in self._epoch(self.testDataIndex,batch_size):
            yield self._returnData(batch)

    def GenerateValidationSampler(self,validation_fraction):
        """

        Returns a new ClinicalDatasetSampler that splits the training set into training and validation sets. This new sampler can be used just like the original except that the testing dataset is now a validation subset of the training data. It accomplishes this task by making a general shallow copy of the class (to avoid copying, for example, the whole dataset)while making deep copies of the information that changes between the validation and test datasets.

        Parameters
        ----------
        validation_fraction : float
            Fraction of the training data to use for validation.

        Returns
        -------
        ClinicalDatasetSampler
            A new ClinicalDatasetSampler with the testing dataset set to a validation subset of the training data.

        """

        new_instance = copy.copy(self)
        new_instance.trainingDataIndex=copy.deepcopy(self.trainingDataIndex)
        new_instance.testDataIndex=copy.deepcopy(self.testDataIndex)
        new_instance.trainingFraction=copy.deepcopy(self.trainingFraction)
        new_instance.trainingFraction = 1.0-validation_fraction

        if self.isConditioned==False:
            new_instance.numTotalSamples=len(self.trainingDataIndex)
            trainingDataIndexShuffled = shuffle(self.trainingDataIndex)
            cutOffVal = int(np.floor(len(trainingDataIndexShuffled)*new_instance.trainingFraction))
            new_instance.trainingDataIndex = trainingDataIndexShuffled[0:cutOffVal]
            new_instance.testDataIndex = trainingDataIndexShuffled[cutOffVal:]
        else:
            new_instance.numTotalSamples=len(self.trainingDataIndex[0])+len(self.trainingDataIndex[1])
            trainingDataIndexShuffled = np.append(*[shuffle(np.array(x)) for x in self.trainingDataIndex])

            has_at_least_one_dx = np.array(np.sum(np.vstack([self.currentClinicalDataset.data.loc[trainingDataIndexShuffled]['has_'+dx] for dx in self.conditionSamplingOnDx]),axis=0),dtype=np.bool)
            dataWithDx = self.currentClinicalDataset.data.loc[trainingDataIndexShuffled].index[has_at_least_one_dx>0]
            dataWithoutDx = self.currentClinicalDataset.data.loc[trainingDataIndexShuffled].index[has_at_least_one_dx==0]
            self.fracWDx = len(dataWithDx)/len(trainingDataIndexShuffled)
            cutOffValWDx = int(np.floor(len(dataWithDx)*new_instance.trainingFraction))
            cutOffValWoDx = int(np.floor(len(dataWithoutDx)*new_instance.trainingFraction))

            new_instance.trainingDataIndex=[dataWithDx[0:cutOffValWDx],dataWithoutDx[0:cutOffValWoDx]]
            new_instance.testDataIndex=[dataWithDx[cutOffValWDx:],dataWithoutDx[cutOffValWoDx:]]
        return new_instance

    def CollapseDataArrays(self,disInds=None,cov_vecs=None,drop_column=False):
        """
        Converts


        Parameters
        ----------
        disInds : type
            Description of parameter `disInds`.
        cov_vecs : type
            Description of parameter `cov_vecs`.
        drop_column : type
            Description of parameter `drop_column`.

        Returns
        -------
        type
            Description of returned object.

        """
        list_of_arrays=[]
        if disInds is not None:
            list_of_arrays+=[disInds]


        if cov_vecs is not None:
            n_cat_vec = [len(self.currentClinicalDataset.catCovConversionDicts[x]) for x in self.includedCovariates]

            for i,n_cat in enumerate(n_cat_vec):
                if torch.is_tensor(cov_vecs[0]):
                    list_of_arrays+=[one_hot(cov_vecs[i],n_cat,dropColumn=drop_column)]
                else:
                    list_of_arrays+=[one_hot_scipy(cov_vecs[i],n_cat,dropColumn=drop_column)]


        list_of_arrays=[self.arrayFunc(x) for x in list_of_arrays]

        if self.returnArrays=='Numpy':
            return np.hstack(list_of_arrays,dtype=np.float64)
        elif self.returnArrays=='Sparse':
            return sparse.hstack(list_of_arrays,format='csr',dtype=np.float64)
        else:
            return torch.cat(list_of_arrays,dim=1,dtype=torch.float32)

    def AddAuxillaryDataset(self,newClinicalDataset):
        """
        Adds another, auxillary ClinicalDataset to the sampler. This way, different clinical datasets for the same sets of patients can be generated in parallel. If activated, this data is returned as the 4th element in the return tuple.

        Parameters
        ----------
        newClinicalDataset : ClinicalDataset
            A ClinicalDataset Class with the same subjects as the current class.

        Returns
        -------
        None

        """
        assert len(self.currentClinicalDataset.data.index.difference(newClinicalDataset.data.index))==0,"Auxillary ClinicalDataset must contain the same samples as the original ClinicalDataset"
        self._returnAuxData=True
        self._auxDataset=newClinicalDataset



    def RemoveAuxillaryDataset(self):
        """
        Removes an auxillary dataset from the sampler

        Returns
        -------
        None

        """
        self._returnAuxData=False
        self._auxDataset=None

class _TorchDatasetWrapper(data.Dataset):

    def __init__(self,clinicalDatasetSampler,sampling_index,batch_size):
        """

        Wrapper for ClinicalData and ClinicalDatasetSampler to allow for rapid subset sampling using PyTorch DataLoader, which allows for multi-threaded loading/queueing of data.

        Parameters
        ----------
        clinicalDatasetSampler : ClinicalDatasetSampler
            ClinicalDatasetSampler to be wrapped
        sampling_index : Iterable or list of Iterables
            Index of subjects from ClincalDataset to sample from. Use ClinicalDatasetSampler.trainingDataIndex or ClinicalDatasetSampler.testingDataIndex
        batch_size : int
            Batch size for the sampler.

        Returns
        -------
        None

        """


        self.clinicalDatasetSampler = clinicalDatasetSampler
        self.sampling_index=sampling_index


        if self.clinicalDatasetSampler.isConditioned:
            #shuffle the order of the dataset at the start of every epoch
            self.sampling_index[0]=shuffle(self.sampling_index[0])
            self.sampling_index[1]=shuffle(self.sampling_index[1])
            self.totalNumBatches,leftover = divmod((len(self.sampling_index[0])+len(self.sampling_index[1])),batch_size)
            assert self.totalNumBatches <= len(sampling_index[0]), "Batch size too small. Cannot ensure at least one positive example per batch."
            #first shuffle the data

            self.hasDxSplits=self.clinicalDatasetSampler._indexSplit(len(self.sampling_index[0]),self.totalNumBatches)
            self.noDxSplits=self.clinicalDatasetSampler._indexSplit(len(self.sampling_index[1]),self.totalNumBatches)
            if self.totalNumBatches == 0:
                self.totalNumBatches+=1

        else:
            self.sampling_index=shuffle(self.sampling_index)
            self.totalNumBatches,leftover = divmod(len(self.sampling_index),batch_size)
            self.splits = self.clinicalDatasetSampler._indexSplit(len(self.sampling_index),self.totalNumBatches)
            if self.totalNumBatches == 0:
                self.totalNumBatches+=1

    def __len__(self):
        return self.totalNumBatches

    def shuffle_index(self):
        if self.clinicalDatasetSampler.isConditioned:
            self.sampling_index[0]=shuffle(self.sampling_index[0])
            self.sampling_index[1]=shuffle(self.sampling_index[1])
        else:
            self.sampling_index=shuffle(self.sampling_index)



    def __getitem__(self,index):
        if self.clinicalDatasetSampler.isConditioned:
            hasDxSubset = self.sampling_index[0][self.hasDxSplits[index]:self.hasDxSplits[index+1]]
            noDxSubset =  self.sampling_index[1][self.noDxSplits[index]:self.noDxSplits[index+1]]
            batchIndex = np.concatenate((hasDxSubset,noDxSubset))

        else:
            batchIndex=self.sampling_index[self.splits[index]:self.splits[index+1]]

        return self.clinicalDatasetSampler._returnData(batchIndex)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri Jul 12 11:26:37 2019

@author: davidblair
"""
import torch
import torch.nn as nn
from typing import Iterable
from vlpi.model.FCLayers import FCLayers,FCLayers_Monotonic
from vlpi.utils.LinearPositiveWeights import LinearPositiveWeights
from vlpi.utils.UtilityFunctions import build_onehot_arrays,random_catcov


class LinearDecoder_Monotonic(nn.Module):


    def __init__(self,nInputDim:int,nCatCovList:Iterable[int],nOutputDim:int,dropLinearCovariateColumn:bool,includeBias:bool):
        """
        Monotonic linear (stritly positive weights) decoder network mapping nInputDim +len(nCatCovList)
        inputs to nOutputDim nodes.

        dropLinearCovariateColumn: whether to drop one value from each categorical covariate
        includeBias: whether or not to include a bias term (intercept) into the network

        """

        super(LinearDecoder_Monotonic, self).__init__()
        self.nCatCovList=nCatCovList
        self.nOutputDim=nOutputDim
        self.dropLinearCovariateColumn=dropLinearCovariateColumn
        self.nInputDim=nInputDim
        if len(nCatCovList)>0:
            if dropLinearCovariateColumn:
                self.nInputDimCov=sum(nCatCovList)-len(nCatCovList)
            else:
                self.nInputDimCov=sum(nCatCovList)
            self.linear_cov = nn.Linear(self.nInputDimCov,nOutputDim,bias=False)
        self.linear_latent = LinearPositiveWeights(self.nInputDim,nOutputDim,bias=includeBias)



    def forward(self,latent_var,*cat_cov_list:int):

        one_hot_cat_list = build_onehot_arrays(cat_cov_list,self.nCatCovList,self.dropLinearCovariateColumn)
        latent_component = self.linear_latent(latent_var)
        if len(one_hot_cat_list)>0:
            return self.linear_cov(torch.cat((*one_hot_cat_list,),dim=-1))+latent_component
        else:
            return latent_component


class LinearDecoder(nn.Module):
    def __init__(self,nInputDim:int,nCatCovList:Iterable[int],nOutputDim:int,dropLinearCovariateColumn:bool,includeBias:bool):
        """
        Linear decoder network mapping nInputDim +len(nCatCovList) inputs to
        nOutputDim nodes.

        dropLinearCovariateColumn: whether to drop one value from each categorical covariate
        includeBias: whether or not to include a bias term (intercept) into the network

        """
        super(LinearDecoder, self).__init__()
        self.nCatCovList=nCatCovList
        self.nOutputDim=nOutputDim
        self.dropLinearCovariateColumn=dropLinearCovariateColumn
        self.nInputDim=nInputDim
        if len(nCatCovList)>0:
            if dropLinearCovariateColumn:
                self.nInputDimCov=sum(nCatCovList)-len(nCatCovList)
            else:
                self.nInputDimCov=sum(nCatCovList)
            self.linear_cov = nn.Linear(self.nInputDimCov,nOutputDim,bias=False)
        self.linear_latent = nn.Linear(self.nInputDim,nOutputDim,bias=includeBias)


    def forward(self,latent_var,*cat_cov_list:int):
        one_hot_cat_list = build_onehot_arrays(cat_cov_list,self.nCatCovList,self.dropLinearCovariateColumn)
        if len(one_hot_cat_list)>0:
            return self.linear_cov(torch.cat((*one_hot_cat_list,),dim=-1))+self.linear_latent(latent_var)
        else:
            return self.linear_latent(latent_var)

class NonlinearMLPDecoder(nn.Module):
    def __init__(self,nInputDim:int,nCatCovList:Iterable[int],nOutputDim:int,dropLinearCovariateColumn:bool,coupleCovariates:bool,n_layers:int=2,n_hidden:int=128,dropout_rate:float=0.2,use_batch_norm:bool=True):
        """
        Nonlinear MLP decoder network mapping nInputDim +len(nCatCovList) inputs to
        nOutputDim nodes.
        """
        super(NonlinearMLPDecoder, self).__init__()
        self.nCatCovList=nCatCovList
        self.nOutputDim=nOutputDim
        self.dropLinearCovariateColumn=dropLinearCovariateColumn
        self.nInputDim=nInputDim
        self.coupleCovariates = coupleCovariates


        if self.coupleCovariates==True:
            self.nonlinear = FCLayers(n_in=self.nInputDim,n_out=n_hidden,n_cat_list=nCatCovList,n_layers=n_layers,n_hidden=n_hidden,dropout_rate=dropout_rate,use_batch_norm=use_batch_norm)
            self.output_layer =  nn.Linear(n_hidden, self.nOutputDim, bias = True)
        else:
            self.nonlinear_latent = FCLayers(n_in=self.nInputDim,n_out=n_hidden,n_cat_list=[],n_layers=n_layers,n_hidden=n_hidden,dropout_rate=dropout_rate,use_batch_norm=use_batch_norm)
            self.output_layer =  nn.Linear(n_hidden, self.nOutputDim,bias=True)
            if len(nCatCovList)>0:
                if dropLinearCovariateColumn:
                    self.nInputDimCov=sum(nCatCovList)-len(nCatCovList)
                else:
                    self.nInputDimCov=sum(nCatCovList)
                self.linear_cov = nn.Linear(self.nInputDimCov,nOutputDim,bias=False)


    def forward(self,latent_var,*cat_cov_list:Iterable[int]):
        if self.coupleCovariates==True:
            return self.output_layer(self.nonlinear(latent_var,*cat_cov_list))
        else:
            one_hot_cat_list = build_onehot_arrays(cat_cov_list,self.nCatCovList,self.dropLinearCovariateColumn)
            if len(one_hot_cat_list)>0:
                return self.output_layer(self.nonlinear_latent(latent_var,[]))+self.linear_cov(torch.cat((*one_hot_cat_list,),dim=-1))
            else:
                return self.output_layer(self.nonlinear_latent(latent_var,[]))


class NonlinearMLPDecoder_Monotonic(nn.Module):
    def __init__(self,nInputDim:int,nCatCovList:Iterable[int],nOutputDim:int,dropLinearCovariateColumn:bool,n_layers:int=2,n_hidden:int=128,dropout_rate:float=0.2,use_batch_norm:bool=False):
        """
        Nonlinear, monotonic decoder network mapping nInputDim +len(nCatCovList) inputs to
        nOutputDim nodes.

        dropLinearCovariateColumn: whether to drop one value from each categorical covariate
        includeBias: whether or not to include a bias term (intercept) into the network

        """

        super(NonlinearMLPDecoder_Monotonic, self).__init__()
        self.nCatCovList=nCatCovList
        self.nOutputDim=nOutputDim
        self.dropLinearCovariateColumn=dropLinearCovariateColumn
        self.nInputDim=nInputDim

        if len(nCatCovList)>0:
            if dropLinearCovariateColumn:
                self.nInputDimCov=sum(nCatCovList)-len(nCatCovList)
            else:
                self.nInputDimCov=sum(nCatCovList)
            self.linear_cov = nn.Linear(self.nInputDimCov,nOutputDim,bias=False)

        self.nonlinear_pos_latent = FCLayers_Monotonic(n_in=self.nInputDim,n_out=n_hidden,n_cat_list=[],n_layers=n_layers,n_hidden=n_hidden,dropout_rate=dropout_rate)
        self.output_layer =  LinearPositiveWeights(n_hidden, self.nOutputDim,bias=True)

    def forward(self,latent_var,*cat_cov_list:Iterable[int]):
        one_hot_cat_list = build_onehot_arrays(cat_cov_list,self.nCatCovList,self.dropLinearCovariateColumn)
        if len(one_hot_cat_list)>0:
            return self.output_layer(self.nonlinear_pos_latent(latent_var,[]))+self.linear_cov(torch.cat((*one_hot_cat_list,),dim=-1))
        else:
            return self.output_layer(self.nonlinear_pos_latent(latent_var,[]))



import numpy as np
import torch
import string
from vlpi.data.ClinicalDataset import ClinicalDataset,ClinicalDatasetSampler
from vlpi.data.ClinicalDataSimulator import ClinicalDataSimulator
from vlpi.vLPI import vLPI
torch.manual_seed(1023)

## set the simulation parameters

numberOfSamples=50000
numberOfSymptoms=20

rareDiseaseFrequency=0.001
numLatentPhenotypes=2

## simulate the data
simulator = ClinicalDataSimulator(numberOfSymptoms,numLatentPhenotypes,rareDiseaseFrequency)
simulatedData=simulator.GenerateClinicalData(numberOfSamples)

## load the data into a ClinicalDataset class
clinicalData = ClinicalDataset()

#change the dataset to have only 20 symptoms (defaults to full ICD10-CM codebook), named after the first 20 letters
allICDCodes = list(clinicalData.dxCodeToDataIndexMap.keys())
symptomConversionMap=dict(zip(allICDCodes[0:numberOfSymptoms],string.ascii_uppercase[0:numberOfSymptoms]))
clinicalData.ConstructNewDataArray(symptomConversionMap)

# now load into the data structure. Note, when using in this manner, it's the users responsibility to make sure that the input data columns match the data columns of the ClinicalDataset.
clinicalData.LoadFromArrays(simulatedData['incidence_data'],simulatedData['covariate_data'],[],catCovDicts=None, arrayType = 'Torch')

## Now load the ClincalDataset into ClinicalDatasetSampler
training_data_fraction=0.75
sampler = ClinicalDatasetSampler(clinicalData,training_data_fraction,returnArrays='Torch')

## Intantation the models
infNumberOfLatentPhenotypes=10
vlpiModel= vLPI(sampler,infNumberOfLatentPhenotypes)

inference_output = vlpiModel.FitModel(batch_size=1000,errorTol=(1.0/numberOfSamples))
vlpiModel.PackageModel('ExampleModel.pth')

inferredCrypticPhenotypes=vlpiModel.ComputeEmbeddings((simulatedData['incidence_data'],simulatedData['covariate_data']))
riskFunction=vlpiModel.ReturnComponents()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Jul  9 08:42:18 2019

@author: davidblair
"""

from unidecode import unidecode
import pkg_resources
import string
import pickle
import pandas as pd

ICD_PATH = pkg_resources.resource_filename('vlpi', 'data/ICDData/')


class ICDCode:
    def __init__(self, code, associated_string,is_terminal, parent_code=None):
        """


        Parameters
        ----------
        code : str
            ICD10 code string.
        associated_string : str
            String defining code in codebook.
        is_terminal : bool
            Indicates whether code is terminal (no children).
        parent_code : bool, optional
            Indicates if code is parent. The default is None.

        Raises
        ------
        ValueError
            If unable to add child codes to known parent code.

        Returns
        -------
        None.

        """


        self.code = code
        self.associated_string = associated_string
        self.parent_code = parent_code
        self.child_codes = []
        self.is_terminal=is_terminal
        self.parent_code = parent_code
        if parent_code is not None:
            if parent_code.is_terminal==False:
                parent_code.child_codes.append(self)
            else:
                raise ValueError("Attempting to add children to terminal node: child = {}, parent = {}".format(code,parent_code.code))

    def __str__(self):
        return self.associated_string

    def __repr__(self):
        return self.code


class ICDUtilities:

    def _convertToUnicode(self,byteString):
        return unidecode(str(byteString,"ISO-8859-1"))

    def _lettersToInt(self,letter,base):
        list(string.ascii_uppercase).index(letter)*base

    def _convertCodeToIntVal(self,code):
        intVal=0
        for base,letter in enumerate(code[::-1]):
            try:
                intVal+=int(letter)*10**(base)
            except ValueError:
                intVal+=int(list(string.ascii_uppercase).index(letter))*10**(base)
        return intVal


    def _findParentInList(self,code,parentList):

        while len(parentList) > 0:
            if parentList[-1] in code:
                return parentList
            else:
                parentList.pop()
        return parentList

    def ReturnCodeObject(self,code):
        """
        Returns full code object (not just string) for a given code.

        Parameters
        ----------
        code : str
            ICD10 code string.

        Returns
        -------
        ICDCode
            ICD10 code class for input string.

        """

        if code in self.setOfUnusableCodes:
            return self.UnusableICDCodes[self.unusableCodeToIndexMap[code.replace('.','')]]
        else:
            return self.UsableICDCodes[self.usableCodeToIndexMap[code.replace('.','')]]


    def DeleteCode(self,del_code):
        """
        Removes the ICD code and all children (if exist) from data structure.

        Parameters
        ----------
        del_code : str
            ICD10 code to delete


        Returns
        -------
        None
        """


        all_del_codes=self._deleteCode(del_code)


        marker_list_usable= [x.code for x in self.UsableICDCodes]
        marker_list_unusable = [x.code for x in self.UnusableICDCodes]


        for del_code in all_del_codes:
            if del_code in self.setOfUnusableCodes:
                self.UnusableICDCodes.pop(marker_list_unusable.index(del_code))
                marker_list_unusable.remove(del_code)
            else:
                self.UsableICDCodes.pop(marker_list_usable.index(del_code))
                marker_list_usable.remove(del_code)

        self.usableCodeToIndexMap=dict(zip(marker_list_usable,range(len(marker_list_usable))))
        self.unusableCodeToIndexMap=dict(zip(marker_list_unusable,range(len(marker_list_unusable))))
        self.setOfUnusableCodes=set(self.unusableCodeToIndexMap.keys())
        self.setOfUsableCodes=set(self.usableCodeToIndexMap.keys())

    def _deleteCode(self,del_code):

        del_code_obj = self.ReturnCodeObject(del_code)
        parent_code = del_code_obj.parent_code
        del_code_list=[del_code]

        if parent_code is not None:
            parent_code.child_codes.remove(del_code_obj)
        if del_code_obj.is_terminal==False:
            for child_code in del_code_obj.child_codes:
                del_code_list+=self._deleteCode(child_code.code)
        return del_code_list



    def AssignCodeToChapter(self,code):
        """
        Returns the chapter heading for input code

        Parameters
        ----------
        code : str
            ICD10 code.

        Returns
        -------
        str
            ICD10 chapter.

        """


        code = code.replace('.','')
        currentCode = self.ReturnCodeObject(code)
        while currentCode.parent_code is not None:
            currentCode = self.ReturnCodeObject(currentCode.parent_code.code)
        return str(currentCode)

    def ReturnSubsumedTerminalCodes(self,parent_code):
        """
        Returns all ICD10 codes that are children of the input code.

        Parameters
        ----------
        parent_code : str
            ICD10 string for parent code. Do not include periods (ie J101 not J10.1)

        Returns
        -------
        terminal_code_list : list
            List of ICD10 codes that are children to parent code.

        """

        all_child_codes = self.ReturnCodeObject(parent_code).child_codes
        terminal_code_list=[]
        for child in all_child_codes:
            if child.is_terminal==True:
                terminal_code_list+=[child.code]
            else:
                terminal_code_list+=self.ReturnSubsumedTerminalCodes(child.code)
        return terminal_code_list



    def __init__(self,useICD10UKBB=False,hierarchyFile=None,chapterFile=None):
        """
        Class that manipulates the ICD10 codebook. It stores the codebook as a simple tree (stored as a list called ICDCodes).


        To initialize the class, expects flat two text files:

            1) ICD10_Chapters.txt--chapter heading for all the codes. Manually constructed.
            2) ICD10 codes and hierarchy: icd10cm_order_2018.txt, https://www.cms.gov/Medicare/Coding/ICD10/2018-ICD-10-CM-and-GEMs.html

        By default, the package ships with 2018 version of ICD10-CM and 2020 version of ICD10 from UK Biobank. You can upgrade to 2019 (or downgrade for that matter) by specifying the path to another ICD10 file. ICD9 codebook could be used instead, but you would need to construct data files that match the structure of the ICD10 files.


        Parameters
        ----------
        useICD10UKBB : bool, optional
            Specifies class to use the UK Biobank version of ICD10 (not ICD10-CM). The default is False.
        hierarchyFile : str, optional
            File path to an alternative code hierarchy file. This may (unlikely) work with other encodings but has not been tested. The default is None.
        chapterFile : str, optional
            File path to an alternative code chapter file (ie main groups of codes). Again, this may work with other encodings but has not been tested. The default is None.

        Raises
        ------
        ValueError
            ValueError raised if unable to parse some line. Prints out the line of interest.

        Returns
        -------
        None.

        """

        if hierarchyFile==None:
            if useICD10UKBB:
                hierarchyFile=ICD_PATH+'icd10_ukbb.txt'
            else:
                hierarchyFile=ICD_PATH+'icd10cm_order_2018.txt'
        if chapterFile==None:
            chapterFile = ICD_PATH+'ICD10_Chapters.txt'

        #quick reference, to avoid having to search full tree for codes.


        #Full list of linked codes
        self.UsableICDCodes=[]
        self.usableCodeToIndexMap={}
        self.setOfUsableCodes=set()

        self.UnusableICDCodes=[]
        self.unusableCodeToIndexMap={}
        self.setOfUnusableCodes=set()
        #first load the chapters
        chapter_breakpoints=[]
        chapter_list=[]
        currentUsableCodeCount = 0
        currentUnusableCodeCount = 0
        with open(chapterFile,'rb') as f:
            f.readline()
            for line in f:
                line=self._convertToUnicode(line)
                line=line.strip('\n').split('\t')
                self.UnusableICDCodes+=[ICDCode('Chapter_'+line[0],line[2],False)]
                start,stop = line[1].split('-')
                chapter_breakpoints+=[self._convertCodeToIntVal(stop[0:3])]
                chapter_list+=['Chapter_'+line[0]]
                self.unusableCodeToIndexMap['Chapter_'+line[0]]=currentUnusableCodeCount
                self.setOfUnusableCodes.add('Chapter_'+line[0])
                currentUnusableCodeCount+=1
        #now load hierarchy file

        with open(hierarchyFile,'rb') as f:
            currentParentList = []
            for line in f:
                line=self._convertToUnicode(line)
                parsedLine=[]
                parsedLine+=[line[0:6].strip()]
                parsedLine+=[line[6:14].strip()]
                parsedLine+=[line[14:16].strip()]
                parsedLine+=[line[77:].strip()]


                currentParentList = self._findParentInList(parsedLine[1],currentParentList)

                if len(currentParentList) == 0:
                    intVal = self._convertCodeToIntVal(parsedLine[1][0:3])
                    try:
                        icd_chapter = chapter_list[next(x[0] for x in enumerate(chapter_breakpoints) if intVal <= x[1])]
                    except StopIteration:
                        raise ValueError('{}'.format(parsedLine[1]))
                    currentParentList +=[icd_chapter]

                if int(parsedLine[2])==1:
                    self.UsableICDCodes+=[ICDCode(parsedLine[1],parsedLine[3],True,self.ReturnCodeObject(currentParentList[-1]))]
                    self.usableCodeToIndexMap[parsedLine[1]]=currentUsableCodeCount
                    self.setOfUsableCodes.add(parsedLine[1])
                    currentUsableCodeCount+=1
                else:
                    self.UnusableICDCodes+=[ICDCode(parsedLine[1],parsedLine[3],False,self.ReturnCodeObject(currentParentList[-1]))]
                    self.unusableCodeToIndexMap[parsedLine[1]]=currentUnusableCodeCount
                    self.setOfUnusableCodes.add(parsedLine[1])
                    currentUnusableCodeCount+=1
                    currentParentList+=[parsedLine[1]]


class ICD10TranslationMap:

    def _buildTranslationTable(self):

        translation_table={'Primary Code':[],'Secondary Code(s)':[],'Relationship':[]}
        for code in self.primaryEncoding.UsableICDCodes:
            #first check if there is 1:1 mapping between codes
            if code.code in self.secondaryEncoding.setOfUsableCodes:
                translation_table['Primary Code']+=[code.code]
                translation_table['Secondary Code(s)']+=[set([code.code])]
                translation_table['Relationship']+=['Direct']
            else:
                parent = code.parent_code
                if parent.code in self.secondaryEncoding.setOfUsableCodes:
                    translation_table['Primary Code']+=[code.code]
                    translation_table['Secondary Code(s)']+=[set([parent.code])]
                    translation_table['Relationship']+=['Parent']
                else:
                    if len(code.child_codes)>0:
                        child_code_names = [x.code for x in code.child_codes]
                        allowed_child_codes = set(child_code_names).intersection(self.secondaryEncoding.setOfUsableCodes)
                        if len(allowed_child_codes)>0:
                            translation_table['Primary Code']+=[code.code]
                            translation_table['Secondary Code(s)']+=[allowed_child_codes]
                            translation_table['Relationship']+=['Child']

        translation_table=pd.DataFrame(translation_table)
        translation_table.set_index('Primary Code',drop=False,inplace=True)
        return translation_table



    def __init__(self,primaryEncoding=None,secondaryEncoding=None):
        """
        Builds translation map between two ICD Utilities with at least some shared codes by taking advantage of shared hierarchical structure.

        If primaryEncoding and secondaryEncoding are unspecified, class creates a map between ICD10-CM and ICD10 (UKBB)

        Parameters
        ----------
        primaryEncoding : ICDUtilities, optional
            First encoding. The default is None.
        secondaryEncoding : ICDUtilities, optional
            Second encoding. The default is None.

        Returns
        -------
        None.

        """

        if (primaryEncoding is not None) or (secondaryEncoding is not None):
            assert (secondaryEncoding is not None) and (secondaryEncoding is not None), "Must specify primary and secondary encoding if providing one or the other."
            self.primaryEncoding=primaryEncoding
            self.secondaryEncoding=secondaryEncoding

        if primaryEncoding is None:
            try:
                with open(ICD_PATH+'icd10cm_to_ukbb.pth','rb') as f:
                    self.EncodingCoversionTable = pickle.load(f)
            except FileNotFoundError:
                self.primaryEncoding = ICDUtilities()
                self.secondaryEncoding=ICDUtilities(useICD10UKBB=True)

                self.EncodingCoversionTable=self._buildTranslationTable()
                self.EncodingCoversionTable.to_pickle(ICD_PATH+'icd10cm_to_ukbb.pth')
        else:
            self.EncodingCoversionTable=self._buildTranslationTable()



    def ReturnConversionSet(self,primaryCode,includeRelationship=False):
        """
        Returns set of codes that represent the mapping of the primary code to the new encoding system.

        Parameters
        ----------
        primaryCode : str
            Diagnostic code to be converted .
        includeRelationship : bool, optional.
            Specicies whether to return the relationship type in addition to code. The default is False.

        Returns
        -------
        set
            Set of codes aligned to the code of interest.

        """

        if includeRelationship:
            look_up=['Secondary Code(s)','Relationship']
        else:
            look_up='Secondary Code(s)'
        try:
            return self.EncodingCoversionTable.loc[primaryCode][look_up]
        except KeyError:
            return set([])



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri Jul 12 11:13:55 2019

@author: davidblair
"""
import torch
import numpy as np
import subprocess
from scipy import sparse
from scipy.stats import fisher_exact as _fisher_exact
from scipy.stats import ttest_ind

def build_onehot_arrays(cat_cov_list,nCatCovList,dropOneColumn):
    """
    Builds one-hot arrays from integer categories. Adapted from scVI.
    cat_cov_list: list of categorical arrays (torch.tensors)
    nCatCovList: list indicating number of categories for each covariate
    dropOneColumn: indicates whether or not to drop one column from the coviariates, eliminates colinearity

    """

    one_hot_cat_list = []  # for generality in this list many indices useless.
    assert len(nCatCovList) == len(cat_cov_list), "number of categorical args provided doesn't match initial params."
    for n_cat, cat in zip(nCatCovList, cat_cov_list):
        assert not (n_cat and cat is None), "category not provided while n_cat != 0 in init. params."
        if n_cat > 1:  # n_cat = 1 will be ignored - no additional information
            if cat.size(1) != n_cat:
                one_hot_cat = one_hot(cat, n_cat,dropOneColumn)
            else:
                one_hot_cat = cat  # cat has already been one_hot encoded
            one_hot_cat_list += [one_hot_cat]
    return one_hot_cat_list

def one_hot(index, n_cat, dropColumn=False):
    """
    expects tensor of shape (n_samples,1), returns one-hot array size
    (n_samples, n_cat).

    Optionally, can drop the first column of the encoding to prevent colinearity
    among the predictors. This is only necessary if you care about the values
    of the inferred parameters during inference.
    """

    onehot = torch.zeros(index.size(0), n_cat, device=index.device)
    onehot.scatter_(1, index.type(torch.long), 1)
    if dropColumn:
        return onehot.type(torch.float32)[:,1:]
    else:
        return onehot.type(torch.float32)

def one_hot_scipy(index,n_cat,dropColumn=False):
    if sparse.issparse(index):
        index = index.toarray().ravel()
    else:
        index=np.array(index).ravel()
    one_hot = sparse.coo_matrix((np.ones(index.shape[0]),(np.arange(index.shape[0]),index)),shape = (index.shape[0],n_cat),dtype=np.float64)
    one_hot = one_hot.tocsr()
    if dropColumn:
        one_hot = one_hot[:,1:]
    return one_hot

def file_len(fname,skip_rows):
    p = subprocess.Popen(['wc', '-l', fname], stdout=subprocess.PIPE,
                                              stderr=subprocess.PIPE)
    result, err = p.communicate()
    if p.returncode != 0:
        raise IOError(err)
    return int(result.strip().split()[0])-skip_rows

def random_catcov(n_cat,n_samples,device):
    samp_params = np.random.dirichlet(np.ones((n_cat)))
    return torch.tensor(np.random.choice(np.arange(n_cat),size=(n_samples,1),p=samp_params),device=device)





def rel_diff(curr,prev):
    return abs((curr - prev) / prev)


def fisher_exact(dataMatrix,incidenceVec):
    """
    dataMatrix: sparse NxK matrix of binary features
    incidenceVec: sparse Nx1 vector of binary labels

    returns: score,pval
    """

    numLabel = incidenceVec.sum()
    whereLabelTrue = incidenceVec.nonzero()[0]
    numTotal = dataMatrix.shape[0]
    scores = np.zeros(dataMatrix.shape[1],dtype=np.float64)
    pvals = np.zeros(dataMatrix.shape[1],dtype=np.float64)

    for feature_i in range(dataMatrix.shape[1]):
        numFeature = dataMatrix[:,feature_i].sum()
        numBoth = dataMatrix[whereLabelTrue,feature_i].sum()
        numFeatureOnly = numFeature - numBoth
        numLabelOnly = numLabel - numBoth
        numNeither = numTotal - numBoth-numLabelOnly-numFeatureOnly
        fisher_test_results = _fisher_exact(np.array([[numBoth,numLabelOnly],[numFeatureOnly,numNeither]]))
        scores[feature_i]=fisher_test_results[0]
        pvals[feature_i]=fisher_test_results[1]
    return scores,pvals


def T_test(dataMatrix,scoreVec):
    scores = np.zeros(dataMatrix.shape[1],dtype=np.float64)
    pvals = np.zeros(dataMatrix.shape[1],dtype=np.float64)
    for feature_i in range(dataMatrix.shape[1]):
        where_nonzero=dataMatrix[:,feature_i].nonzero()[0]
        where_zero = np.setdiff1d(np.arange(dataMatrix.shape[0]),where_nonzero)

        pop_a = scoreVec[where_nonzero]
        pop_b = scoreVec[where_zero]

        stats = ttest_ind(pop_a, pop_b,equal_var=False)
        scores[feature_i]=stats[0]
        pvals[feature_i]=stats[1]
    return scores,pvals



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu Aug 22 09:11:51 2019

@author: davidblair
"""

###

import setuptools
import re

version = re.search(
    '^__version__\s*=\s*"(.*)"',
    open('vlpi/vLPI.py').read(),
    re.M).group(1)



with open("README.md", "r") as fh:
    long_description = fh.read()



setuptools.setup(
    name="vlpi",
    version=version,
    author="David Blair",
    author_email="david.blair@ucsf.edu",
    description="Python implementation of the variational latent phenotype model described in Blair et al..",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/daverblair/vlpi",
    package_data={'vlpi': ['data/ICDData/*.txt']},
    include_package_data=True,
    packages=setuptools.find_packages(),
    install_requires=[
        'torch>=1.5.1',
        'pyro-ppl>=1.3.1',
        'numpy>=1.19.0',
        'pandas>=1.0.5',
        'scipy>=1.5.2',
        'scikit-learn>=0.22.1',
        'typing',
        'unidecode'
        ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri Jul 12 08:59:06 2019

@author: davidblair
"""

import collections
from typing import Iterable
import torch
from torch import nn as nn
from vlpi.utils.UtilityFunctions import one_hot
from vlpi.utils.LinearPositiveWeights import LinearPositiveWeights


class FCLayers_Monotonic(nn.Module):
    """A helper class to build fully-connected layers for a neural network.
    Uses Tanh for non-linearities; dropoutfor convergence.
    Adapted frome the scVI package: https://github.com/YosefLab/scVI.

    :param n_in: The dimensionality of the input data (excluding covariates)
    :param n_out: The dimensionality of the output layer
    :param n_cat_list: A list containing, for each categorical covariate of interest,
                 the number of categories. Each category will be
                 included using a one-hot encoding.
    :param n_layers: The number of fully-connected hidden layers
    :param n_hidden: The number of nodes per hidden layer
    :param dropout_rate: Dropout rate to apply to each of the hidden layers
    Note: there no batch normalization with this module, as this can break the
    positive correlation structure desired by the model. Moreover, we replaced
    the ReLU function with tanh function which ensures monotonicity without erasing
    negative values (which the ReLU will do by design)
    """
    def __init__(self, n_in: int, n_out: int, n_cat_list: Iterable[int] = None,
                 n_layers: int = 2, n_hidden: int = 128, dropout_rate: float = 0.2, use_batch_norm=True):

        super(FCLayers_Monotonic,self).__init__()

        layers_dim = [n_in] + (n_layers - 1) * [n_hidden] + [n_out]

        if n_cat_list is not None:
            # n_cat = 1 will be ignored
            self.n_cat_list = [n_cat if n_cat > 1 else 0 for n_cat in n_cat_list]
        else:
            self.n_cat_list = []

        self.fc_layers = nn.Sequential(collections.OrderedDict(
            [('Layer {}'.format(i), nn.Sequential(
                LinearPositiveWeights(n_in + sum(self.n_cat_list), n_out,bias=True),
                nn.Tanh(),
                nn.Dropout(p=dropout_rate) if dropout_rate > 0 else None))
             for i, (n_in, n_out) in enumerate(zip(layers_dim[:-1], layers_dim[1:]))]))

    def forward(self, x: torch.Tensor, *cat_list: int):
        """Forward computation on ``x``.
        :param x: tensor of values with shape ``(n_samples,n_in)``
        :param cat_list: list of category membership(s) for this sample, provided as (1,n_sample) tensors
        :return: tensor of shape ``(n_out,n_samples)``
        :rtype: :py:class:`torch.Tensor`
        """
        one_hot_cat_list = []  # for generality in this list many indices useless.
        assert len(self.n_cat_list) <= len(cat_list), "nb. categorical args provided doesn't match init. params."
        for n_cat, cat in zip(self.n_cat_list, cat_list):
            assert not (n_cat and cat is None), "cat not provided while n_cat != 0 in init. params."
            if n_cat > 1:  # n_cat = 1 will be ignored - no additional information
                if cat.size(1) != n_cat:
                    one_hot_cat = one_hot(cat, n_cat)
                else:
                    one_hot_cat = cat  # cat has already been one_hot encoded
                one_hot_cat_list += [one_hot_cat]
        for layers in self.fc_layers:
            for layer in layers:
                if layer is not None:
                    if isinstance(layer, LinearPositiveWeights):
                        if x.size(0)!=0:
                            x = torch.cat((x, *one_hot_cat_list), dim=-1)
                        else:
                            x = torch.cat((*one_hot_cat_list,), dim=-1)

                    x = layer(x)
        return x


class FCLayers(nn.Module):
    """A helper class to build fully-connected layers for a neural network.
    Uses ReLU for non-linearities, BatchNorm1d and dropout for convergence.
    Adapted frome the scVI package: https://github.com/YosefLab/scVI.

    :param n_in: The dimensionality of the input data (excluding covariates)
    :param n_out: The dimensionality of the output layer
    :param n_cat_list: A list containing, for each categorical covariate of interest,
                 the number of categories. Each category will be
                 included using a one-hot encoding.
    :param n_layers: The number of fully-connected hidden layers
    :param n_hidden: The number of nodes per hidden layer
    :param dropout_rate: Dropout rate to apply to each of the hidden layers
    :param use_batch_norm: whether or not to use batch normalization to speed convergence
    """


    def __init__(self, n_in: int, n_out: int, n_cat_list: Iterable[int] = None,
                 n_layers: int = 1, n_hidden: int = 128, dropout_rate: float = 0.2, use_batch_norm=True):

        super(FCLayers,self).__init__()

        layers_dim = [n_in] + (n_layers - 1) * [n_hidden] + [n_out]

        if n_cat_list is not None:
            # n_cat = 1 will be ignored
            self.n_cat_list = [n_cat if n_cat > 1 else 0 for n_cat in n_cat_list]
        else:
            self.n_cat_list = []

        self.fc_layers = nn.Sequential(collections.OrderedDict(
            [('Layer {}'.format(i), nn.Sequential(
                nn.Linear(n_in + sum(self.n_cat_list), n_out,bias=True),
                nn.BatchNorm1d(n_out) if use_batch_norm else None,
                nn.ReLU(),
                nn.Dropout(p=dropout_rate) if dropout_rate > 0 else None))
             for i, (n_in, n_out) in enumerate(zip(layers_dim[:-1], layers_dim[1:]))]))

    def forward(self, x: torch.Tensor, *cat_list: int):
        """Forward computation on ``x``.
        :param x: tensor of values with shape ``(n_samples,n_in)``
        :param cat_list: list of category membership(s) for this sample, provided as (1,n_sample) tensors
        :return: tensor of shape ``(n_out,n_samples)``
        :rtype: :py:class:`torch.Tensor`
        """
        one_hot_cat_list = []  # for generality in this list many indices useless.
        assert len(self.n_cat_list) <= len(cat_list), "nb. categorical args provided doesn't match init. params."
        for n_cat, cat in zip(self.n_cat_list, cat_list):
            assert not (n_cat and cat is None), "cat not provided while n_cat != 0 in init. params."
            if n_cat > 1:  # n_cat = 1 will be ignored - no additional information
                if cat.size(1) != n_cat:
                    one_hot_cat = one_hot(cat, n_cat)
                else:
                    one_hot_cat = cat  # cat has already been one_hot encoded
                one_hot_cat_list += [one_hot_cat]
        for layers in self.fc_layers:
            for layer in layers:
                if layer is not None:
                    if isinstance(layer, nn.Linear):
                        if x.size(0)!=0:
                            x = torch.cat((x, *one_hot_cat_list), dim=-1)
                        else:
                            x = torch.cat((*one_hot_cat_list,), dim=-1)
                    x = layer(x)
        return x



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu Dec 12 10:42:20 2019

@author: davidblair
"""

import torch
import torch.nn as nn
import pyro
import pyro.distributions as dist
from pyro.infer import Trace_ELBO
from typing import Iterable
import copy
from vlpi.model.Encoders import MeanScaleEncoder
from vlpi.model.Decoders import LinearDecoder_Monotonic,LinearDecoder,NonlinearMLPDecoder,NonlinearMLPDecoder_Monotonic
from vlpi.utils.UtilityFunctions import random_catcov


class VAE(nn.Module):


    def SwitchDevice(self,new_compute_device):
        """

        Switches model to a different device.

        Parameters
        ----------
        new_compute_device : int or str
            Parameter that specifies the device. Can be integer (GPU) or str ('cpu')

        """
        self.compute_device=new_compute_device
        self.to(self.compute_device)


    def __init__(self,numObsTraits:int, numCatList:Iterable[int],nLatentDim:int,decoderType:str,**kwargs):
        """

        Variational autoencoder used for latent phenotype inference.

        Parameters
        ----------
        numObsTraits : int
            Number of traits or symptoms used in the model.
        numCatList : Iterable[int]
            List containing the number of categories for each categorical covariate.
        nLatentDim : int
            Number of latent dimensions in the model
        decoderType : str
            Type of decoder. Must be one of the following: 'Linear','Linear_Monotonic','Nonlinear','Nonlinear_Monotonic'
        **kwargs : type
            Mutliple kwargs available. Please see source code for details.

        Returns
        -------
        Nonlinear

        """
        super(VAE,self).__init__()
        self.numObsTraits=numObsTraits
        self.nLatentDim = nLatentDim
        self.numCatList=numCatList
        self.decoderType=decoderType

        assert self.decoderType in ['Linear','Linear_Monotonic','Nonlinear','Nonlinear_Monotonic'], "Currently supported decoders for VAE include: 'Linear','Linear_Monotonic','Nonlinear','Nonlinear_Monotonic'"

        allKeywordArgs = list(kwargs.keys())


        if 'linkFunction' not in allKeywordArgs:
            self.linkFunction = lambda x:torch.sigmoid(x)
        else:
            linkFunction = kwargs['linkFunction']
            assert linkFunction in ['Logit','Probit'],"Only Logit and Probit link functions currently supported."

            if linkFunction=='Logit':
                self.linkFunction = lambda x:torch.sigmoid(x)
            else:
                self.linkFunction = lambda x:dist.Normal(torch.tensor(0.0,dtype=torch.float32,device=x.device),torch.tensor(1.0,dtype=torch.float32,device=x.device)).cdf(x)

        if 'computeDevice' not in allKeywordArgs:
            """
            Specifies compute device for model fitting, can also be specified later by
            calling SwitchDevice

            """
            self.compute_device=None
        else:
            self.compute_device=kwargs['computeDevice']


        if 'dropLinearCovariateColumn' not in allKeywordArgs:
            """
            Specifies model to drop one category from each covariate. Defaults to True.
            """
            self.dropLinearCovariateColumn=True
        else:
            self.dropLinearCovariateColumn=kwargs['dropLinearCovariateColumn']
            assert isinstance(self.dropLinearCovariateColumn,bool),"dropLinearCovariateColumn expects boolean value"

        if 'coupleCovariates' not in allKeywordArgs:
            """
            Specifies whether to couple covariates to non-linear MLP network (True), or to model them  using an independent linear network (False). Defaults to True.
            """
            self.coupleCovariates=True
        else:
            self.coupleCovariates=kwargs['coupleCovariates']
            assert isinstance(self.dropLinearCovariateColumn,bool),"coupleCovariates expects boolean value"
            if self.decoderType not in ['Nonlinear']:
                print("Warning: Not fitting Nonlinear model. Coupling covariates has no effect on inference.")


        if 'encoderNetworkHyperparameters' not in allKeywordArgs:
            self.encoderHyperparameters={'n_layers' : 2, 'n_hidden' : 64, 'dropout_rate': 0.0, 'use_batch_norm':True}

        else:
            self.encoderHyperparameters = kwargs['encoderNetworkHyperparameters']
            assert isinstance(self.encoderHyperparameters,dict),"Expects dictionary of encoder hyperparameters"
            assert set(self.encoderHyperparameters.keys())==set(['n_layers','n_hidden','dropout_rate','use_batch_norm']),"Encoder hyperparameters must include: 'n_layers','n_hidden','dropout_rate','use_batch_norm'"

        if 'decoderNetworkHyperparameters' not in allKeywordArgs:
            self.decoderHyperparameters={'n_layers' : 2, 'n_hidden' : 64, 'dropout_rate': 0.0, 'use_batch_norm':True}

        else:
            self.decoderHyperparameters = kwargs['decoderNetworkHyperparameters']
            assert isinstance(self.encoderHyperparameters,dict),"Expects dictionary of decoder hyperparameters"
            assert set(self.encoderHyperparameters.keys())==set(['n_layers','n_hidden','dropout_rate','use_batch_norm']),"Decoder hyperparameters must include: 'n_layers','n_hidden','dropout_rate','use_batch_norm'"

            if self.decoderType not in ['Nonlinear','Nonlinear_Monotonic']:
                print("Warning: Decoder neural network hyperparameters specified for a linear model. Parameters will not be used.")

        if self.dropLinearCovariateColumn:
            self.numCovParam = sum(self.numCatList)-len(self.numCatList)
        else:
            self.numCovParam = sum(self.numCatList)

        self.encoder=MeanScaleEncoder(self.numObsTraits, self.nLatentDim, n_cat_list=self.numCatList, **self.encoderHyperparameters)

        if self.decoderType=='Linear':
            self.decoder = LinearDecoder(self.nLatentDim,self.numCatList,self.numObsTraits,self.dropLinearCovariateColumn,True)
        elif self.decoderType=='Linear_Monotonic':
            self.decoder = LinearDecoder_Monotonic(self.nLatentDim,self.numCatList,self.numObsTraits,self.dropLinearCovariateColumn,True)
        elif self.decoderType == 'Nonlinear':
            self.decoder = NonlinearMLPDecoder(self.nLatentDim,self.numCatList,self.numObsTraits,self.dropLinearCovariateColumn,self.coupleCovariates,**self.decoderHyperparameters)
        else:
            self.decoder = NonlinearMLPDecoder_Monotonic(self.nLatentDim,self.numCatList,self.numObsTraits,self.dropLinearCovariateColumn,**self.decoderHyperparameters)


        if self.compute_device is not None:
            self.SwitchDevice(self.compute_device)
        self.eval()


    def model(self, obs_data=None, cat_cov_list=None,label_dx=None,encoded_data=None,numSamples=None,minibatch_scale=1.0, annealing_factor=1.0):
        if obs_data is not None:
            numSamples=obs_data.shape[0]
        else:
                #assert numSamples is not None, "Must provide either observed data or number of samples to simulate."
            if numSamples is None:
                numSamples=1000
                print('Warning: no arguments were given to VAE.model. This should only be done during debugging.')

            cat_cov_list = []
            for n_cat in self.numCatList:
                cat_cov_list+=[random_catcov(n_cat,numSamples,device=self.compute_device)]

        pyro.module("decoder", self.decoder,update_module_params=False)

        with pyro.poutine.scale(None,minibatch_scale):
            with pyro.plate("latent_pheno_plate",size=numSamples):
                with pyro.poutine.scale(None, annealing_factor):

                    latentPhenotypes=pyro.sample("latentPhenotypes",dist.Normal(torch.zeros(1,self.nLatentDim,dtype=torch.float32,device=self.compute_device),torch.ones(1,self.nLatentDim,dtype=torch.float32,device=self.compute_device)).to_event(1))

                liability_vals = self.decoder.forward(latentPhenotypes,*cat_cov_list)
                latent_dx_prob = self.linkFunction(liability_vals)
                pyro.sample("obsTraitIncidence",dist.Bernoulli(latent_dx_prob).to_event(1),obs=obs_data)


    def guide(self,obs_data=None, cat_cov_list=None,label_dx=None,encoded_data=None,numSamples = None,minibatch_scale=1.0,annealing_factor=1.0):
        if obs_data is not None:
            numSamples=obs_data.shape[0]
        else:
            pass

        pyro.module("encoder", self.encoder,update_module_params=False)


        with pyro.poutine.scale(None,minibatch_scale):
            with pyro.plate("latent_pheno_plate",size=numSamples):
                z_mean,z_std = self.encoder(obs_data,*cat_cov_list)
                with pyro.poutine.scale(None, annealing_factor):
                    pyro.sample("latentPhenotypes", dist.Normal(z_mean, z_std).to_event(1))



    def ComputeELBOPerDatum(self,obs_dis_array,cat_cov_list,num_particles=10):
        """

        Computes the evidence lower bound (ELBO) for each observation in the dataset.

        Parameters
        ----------
        obs_dis_array : torch.tensor
            Binary array of observed data.
        cat_cov_list : list of torch.tensor
            List of categorical covariate values for the dataset
        num_particles : int
            Number of particles (samples) used to approximate the ELBOs

        Returns
        -------
        torch.tensor
            Per-datum ELBOs

        """
        elboFunc = Trace_ELBO(num_particles=num_particles)
        elboVec = torch.zeros(obs_dis_array.shape[0],dtype=torch.float32,device=self.compute_device)

        for model_trace, guide_trace in elboFunc._get_traces(self.model, self.guide,(obs_dis_array,cat_cov_list),{}):
            elboVec+=model_trace.nodes['obsTraitIncidence']['log_prob'].detach()/num_particles
            elboVec+=model_trace.nodes['latentPhenotypes']['log_prob'].detach()/num_particles
            elboVec-=guide_trace.nodes['latentPhenotypes']['log_prob'].detach()/num_particles
        return elboVec.reshape(elboVec.shape[0],1)

    def PredictLatentPhenotypes(self,obs_dis_array,cat_cov_list,returnScale=False,num_particles=10):
        """
        Produces estimates of posterior mean (and scale) of latent phenotypes given some set of observations.

        Parameters
        ----------
        obs_dis_array : torch.tensor
            Binary array of observed data.
        cat_cov_list : list of torch.tensor
            List of categorical covariate values for the dataset
        returnScale : bool
            Whether or not to include scale of posterior as well.
        num_particles : int
            Number of particles (samples) used to approximate the ELBOs

        Returns
        -------
        torch.tensor
            2-D (or 3-D if including scale) array of posterior distribution means (plus scale)

        """

        z_mean,z_scale = self.encoder(obs_dis_array,*cat_cov_list)
        if returnScale:
            return z_mean,z_scale
        else:
            return z_mean


    def PackageCurrentState(self):
        """
        Packages the model state dict as a dictionary.

        Returns
        -------
        dict
            Model state dict.

        """
        packaged_model_state={}
        packaged_model_state['model_state'] = copy.deepcopy(self.state_dict(keep_vars=True))
        return packaged_model_state

    def LoadPriorState(self,prior_model_state):
        """
        Loads model state from dictionary

        Parameters
        ----------
        prior_model_state : dict
            Dictionary of model state produced by PackageCurrentState

        Returns
        -------
        None

        """
        self.load_state_dict(prior_model_state['model_state'],strict=True)
